# -*- coding:utf-8 -*-

import os
import stat
import re
import json
import time
import logging
import threading
import concurrent.futures
from math import ceil
from shutil import get_terminal_size, register_unpack_format
from typing import Optional

from .utils import confirm
from .common import Color, Fx, Symbol
from .common.str_utils import shorten, get_file_icon, adjudgment_type
from .common.str_table import Table


Log = logging.getLogger(__name__)


class CodeCounterError(Exception):
    """CodeCounter error class."""

    pass


class CodeCounter(object):
    """Class of statistical code.

    Attributes:
        Absolute_Rules (dict): Precompiled rules.
        Suffix_Type (dict): Supported file suffix dictionary.
        Special_Name (dict): Type dict of special file name.
        level_color (list): Color list. The levels are calibrated by
            subscript, and the codes of different levels are colored
            when the results are output.
        symbol (dict):
        _support_format (list):
    """

    # The default rule is to count only files. Ignore all video, audio, fonts, binaries.
    Absolute_Rules: list[dict] = [
        # Exclude `.git` folder.
        {"pattern": re.compile(r"\.git$|\.git\/"), "include": False},
        {
            # Exclude all picture formats.
            "pattern": re.compile(
                r"\.xbm$|\.tif$|\.pjp$|\.svgz$|\.jpg$|\.jpeg$|\.ico$|\.icns$|\.tiff$|\.gif$|\.svg$|\.jfif$|\.webp$|\.png$|\.bmp$|\.jpeg$|\.avif$",
                re.I,
            ),
            "include": False,
        },
        {
            # Exclude all video formats.
            "pattern": re.compile(
                r"\.avi$|\.rmvb$|\.rm$|\.asf$|\.divx$|\.mpg$|\.mpeg$|\.mpe$|\.wmv$|\.mp4$|\.mkv$|\.vob$",
                re.I,
            ),
            "include": False,
        },
        {
            # Exclude all audio frequency formats.
            "pattern": re.compile(
                r"\.mp3$|\.wma$|\.mid[i]?$|\.mpeg$|\.cda$|\.wav$|\.ape$|\.flac$|\.aiff$|\.au$",
                re.I,
            ),
            "include": False,
        },
        {
            # Exclude all font formats.
            "pattern": re.compile(
                r"\.otf$|\.woff$|\.woff2$|\.ttf$|\.eot$",
                re.I,
            ),
            "include": False,
        },
        {
            # Exclude some binary file.
            "pattern": re.compile(
                r"\.exe$|\.bin$",
                re.I,
            ),
            "include": False,
        },
    ]

    # Colors displayed for different code quantities.
    level_color: list[str] = [
        "",
        Color.fg("#EBCB8C"),  # yellow
        Color.fg("#FF6347"),  # tomato
        Color.fg("#C71585"),  # middle violet red
        Color.fg("#87CEFA"),  # skyblue
    ]

    # Symbol corresponds to the desired color.
    symbol_color: dict = {"+": Color.fg("#98FB98"), "-": Color.fg("#FF6347")}

    # Supported output format.
    _support_format: list = ["table", "simple"]

    _Lock = threading.Lock()

    def __init__(
        self,
        count_path: str = os.getcwd(),
        use_ignore: bool = True,
        result_saved_path: str = "",
        result_format: str = "table",
        use_icon: bool = False,
    ) -> None:
        """
        Args:
            count_path (str, optional): The path of needed count. Defaults to os.getcwd().
            use_ignore (bool, optional): Whether detect `.gitignore` file. Defaults to True.
            result_saved_path (str, optional): Result save path. Defaults to "".
            result_format (str, optional): Output format string. Defaults to "table".
            use_icon (bool, optional): Whether output with icon. Defaults to False.

        Raises:
            CodeCounterError: when format string not right.
        """
        super(CodeCounter, self).__init__()

        # Store the rules obtained after processing.
        self.Rules: list = []

        self.count_path = count_path
        self.use_ignore = use_ignore
        self.result_saved_path = result_saved_path
        if result_format not in self._support_format:
            raise CodeCounterError(
                "Unsupported format, choice in {0}".format(self._support_format)
            )
        self.result_format = result_format
        self.use_icon = use_icon
        self.rune = Symbol.rune["bold"]

    def process_gitignore(self, root: str) -> None:
        """Process `.gitignore` files and add matching rules.

        Args:
            root (str): Absolute or relative path to the directory.
            files (list): The list of all file names under the `root` path.
        """

        root = root.replace("\\", "/")  # Unified symbol.
        ignore_path = os.path.join(root, ".gitignore")
        try:
            with open(ignore_path) as f:
                ignore_content = filter(
                    # Filter out comment lines.
                    lambda x: x and not x.startswith("#"),
                    map(
                        # Filter out white space lines.
                        # Replace `\` to `/` for windows.
                        lambda x: x.strip().replace("\\", "/"),
                        # Read the file and split the lines.
                        f.read().split("\n"),
                    ),
                )
        except PermissionError:
            if confirm(
                "Can't read {0}, whether get jurisdiction[y/n]:".format(ignore_path)
            ):
                os.chmod(ignore_path, stat.S_IXGRP)
                os.chmod(ignore_path, stat.S_IWGRP)
                self.process_gitignore(root)
        except Exception as e:
            print("Read gitignore error: {0}".format(e))
        else:
            for item in ignore_content:
                is_negative = item[0] == "!"
                if is_negative:
                    item = item[1:]

                slash_index = item.find("/")
                if slash_index == 0:
                    item = root + item
                elif slash_index == -1 or slash_index == len(item) - 1:
                    item = "/".join([root, "**", item])
                else:
                    item = "/".join([root, item])

                item = re.sub(r"([\{\}\(\)\+\.\^\$\|])", r"\1", item)  # escape char
                item = re.sub(r"(^|[^\\])\?", ".", item)
                item = re.sub(r"\/\*\*", "([\\\\/][^\\\\/]+)?", item)  # /**
                item = re.sub(r"\*\*\/", "([^\\\\/]+[\\\\/])?", item)  # **/
                item = re.sub(r"\*", "([^\\\\/]+)", item)  # for `*`
                item = re.sub(r"\?", "*", item)  # for `?``
                item = re.sub(r"([^\/])$", r"\1(([\\\\/].*)|$)", item)
                item = re.sub(r"\/$", "(([\\\\/].*)|$)", item)  # for trialing with `/`
                self.Rules.append({"pattern": re.compile(item), "include": is_negative})

    def matching(self, full_path: str) -> bool:
        """Matching rules.

        Judge whether it is the required file according to the rule matching path.
        Returns `True` if the file not needs to be ignored, or `False` if needs.

        Args:
            full_path (str): File full path for matching.
        """

        # Precompiled rules have the highest priority.
        if list(
            filter(lambda rule: rule["pattern"].search(full_path), self.Absolute_Rules)
        ):
            return False

        # Matching the generated rules.
        res = list(filter(lambda rule: rule["pattern"].search(full_path), self.Rules))
        if not res:
            return True
        else:
            # If multiple rules match successfully, we think the last rule added has
            # the highest priority. Or if just one, this no problem also.
            return res[-1]["include"]
            # selected_rule = max(res, key=lambda rule: len(str(rule["pattern"])))

    def _sub_count(self, root: str, files: list) -> tuple:
        """Process handle use by `self.count`."""
        result = {}  # type: dict[str,dict]
        valid_counter = invalid_counter = 0
        invalid_list = []
        total_size = 0

        for file in files:
            full_path = os.path.join(root, file)
            is_effective = self.matching(full_path)
            if is_effective:
                try:
                    # Try read size of the valid file. Then do sum calc.
                    size_ = os.path.getsize(full_path)
                    total_size += size_
                except:
                    Log.debug(f"Can't read size of '{full_path}'")

                # Get file type.
                type_ = adjudgment_type(file, original=True)
                try:
                    with open(full_path) as f:
                        count = len(f.read().split("\n"))
                except Exception:
                    invalid_counter += 1
                    invalid_list.append(file)
                    continue
                else:
                    # Superposition.
                    if result.get(type_, None) is None:
                        result[type_] = {"files": 1, "lines": count}
                    else:
                        result[type_]["files"] += 1
                        result[type_]["lines"] += count
                    valid_counter += 1
                finally:
                    pass

        return result, total_size, valid_counter, invalid_counter, invalid_list

    @staticmethod
    def _walk_err_callback(e):
        """Handle of processing walk error."""
        print("Walk error: {0}".format(e))
        raise SystemExit(0)

    def count(
        self, root_path: str, use_ignore: bool = True, progress: bool = True
    ) -> tuple[dict, list, int]:
        """Statistics file and returns the result dictionary for Python3.

        Args:
            root_path (str): The path is walk needed.
            use_ignore (bool): Whether ignore files in `.gitignore`. Defaults to True.
            progress (bool): Whether show processing. Defaults to True.

        Return:
            result (dict): Dictionary containing statistical results.
            invalid_list (list): invalid file list.
            total_size (int): the sum size of all valid files.

        >>> result = {
        ...     'py': {
        ...         'files': 5,
        ...         'lines': 2124,
        ...     }
        ... }
        >>> CodeCounter().count('~/.config', use_ignore=True)
        """

        if progress:
            width, _msg = self._get_ready()

        result = {}  # type: dict[str,dict]
        data_count = [0, 0, 0]  # [total_size, valid_count, invalid_count]
        invalid_list = []

        def _callback(r):
            (
                _result,
                _total_size,
                _valid_counter,
                _invalid_counter,
                _invalid_list,
            ) = r.result()

            with self._Lock:
                for key, values in _result.items():
                    if result.get(key, None) is None:
                        result[key] = values
                    else:
                        result[key]["files"] += _result[key]["files"]
                        result[key]["lines"] += _result[key]["lines"]

                data_count[0] += _total_size
                data_count[1] += _valid_counter
                data_count[2] += _invalid_counter
                invalid_list.extend(_invalid_list)

            if progress:
                pass
                # print(
                #     _msg.format(data_count[1], data_count[2]),
                #     end="",
                # )

        cpu: int = os.cpu_count() or 1
        max_queue = cpu * 200
        print("Detect CPU count: {0}, start record ...".format(cpu))

        with concurrent.futures.ProcessPoolExecutor() as pool:
            for root, _, files in os.walk(root_path, onerror=self._walk_err_callback):

                # First judge whether the directory is valid. Invalid directories
                # do not traverse files.
                is_effective_dir = self.matching(root)
                if not is_effective_dir:
                    continue

                # Process .gitignore file, add custom rules.
                if use_ignore and ".gitignore" in files:
                    self.process_gitignore(root)

                if not files:
                    continue

                # print(pool._queue_count)
                if len(files) >= 15 and pool._queue_count < max_queue:
                    # Calling process.
                    future_result = pool.submit(self._sub_count, root, files)
                    future_result.add_done_callback(_callback)
                else:
                    for file in files:
                        full_path = os.path.join(root, file)
                        is_effective = self.matching(full_path)
                        if is_effective:
                            try:
                                # Try read size of the valid file. Then do sum calc.
                                size_ = os.path.getsize(full_path)
                                with self._Lock:
                                    data_count[0] += size_
                            except:
                                Log.debug(f"Can't read size of '{full_path}'")

                            # Get file type.
                            type_ = adjudgment_type(file, original=True)
                            try:
                                with open(full_path) as f:
                                    count = len(f.read().split("\n"))
                            except Exception:
                                with self._Lock:
                                    data_count[2] += 1
                                    invalid_list.append(file)
                                continue
                            else:
                                # Superposition.
                                with self._Lock:
                                    if result.get(type_, None) is None:
                                        result[type_] = {"files": 1, "lines": count}
                                    else:
                                        result[type_]["files"] += 1
                                        result[type_]["lines"] += count
                                    data_count[1] += 1
                            finally:
                                if progress:
                                    # print(
                                    #     _msg.format(data_count[1], data_count[2]),
                                    #     end="",
                                    # )
                                    pass
            print("\nPlease wait calculate ...")

        if progress:
            print("")
        return result, invalid_list, data_count[0]

    def _get_ready(self) -> tuple[int, str]:
        width, _ = get_terminal_size()
        if width > 55:
            _msg = "\rValid files found: {:,}, Invalid files found: {:,}"
        else:
            _msg = "\r:: [{:,} | {:,}]"
        return width, _msg

    def _get_file_path(self, root_path: str) -> str:
        file_name: str = (
            root_path.replace("/", "_").replace("\\", "_").replace(".", "_")
        )
        return os.path.join(self.result_saved_path, file_name)

    def load_recorded_result(self, root_path: str) -> Optional[dict]:
        """Load count result."""
        file_path = self._get_file_path(root_path)
        try:
            with open(file_path) as rf:
                res = json.load(rf)
        except Exception:
            return None
        else:
            return res

    def save_result(self, result: dict, root_path: str) -> bool:
        """Save count result.

        Generate name according to `root_path`, then try save the record
        result to [`TOOLS_HOME`/Counter].

        Args:
            result (dict): Statistical results.
            root_path (str): Traversal directory.

        Return:
            (bool): Whether saving successful.
        """

        file_path = self._get_file_path(root_path)
        # ensure_path(CodeCounter.Result_Saved_Path)
        try:
            with open(file_path, "w" if os.path.isfile(file_path) else "x") as wf:
                json.dump(result, wf, indent=2)
        except Exception:
            return False
        else:
            return True

    @classmethod
    def color_index(cls, _count: int) -> int:
        _index = len(str(_count // 1000))
        if _index > len(cls.level_color):
            return -1
        else:
            return _index - 1

    def format_print(self, new: dict, old: Optional[dict] = None) -> None:
        """Print result with color and diff.

        If the console width is not enough, the output is simple.

        Args:
            new (dict): Current statistical results.
            old (dict|None): The results saved in the past may not exist.
        """

        result_format: str = self.result_format
        needed_width: int = 67
        width, _ = get_terminal_size()
        if result_format == "simple" or width < needed_width:
            for key, value in new.items():
                line = "::{} -> {:,} | {:,}".format(key, value["files"], value["lines"])
                print(line)
            return None

        elif result_format == "table":
            title = "[Code Counter Result]"
            header = ["Language", "Files", "Code lines"]

            # Print full time.
            print(time.strftime("%H:%M:%S %a %Y-%m-%d %Z", time.localtime()))

            # Diff
            sum_ = 0
            additions = 0
            deletions = 0
            tb_data = []
            for key, value in new.items():
                if self.use_icon:
                    key_display_str = "{0} {1}".format(get_file_icon(key), key)
                else:
                    key_display_str = key

                # Processing too long name.
                key_display_str = shorten(key_display_str, 20, front=False)

                # Set color.
                lines_color = self.level_color[self.color_index(value["lines"])]

                # Compare change.
                if isinstance(old, dict) and old.get(key, None) is not None:
                    old_files = old.get(key).get("files", None)
                    old_lines = old.get(key).get("lines", None)

                    if old_files and old_files != value["files"]:
                        files_change = "{:+}".format(value["files"] - old_files)
                        files_symbol = files_change[0]
                    else:
                        files_symbol = files_change = ""

                    if old_lines and old_lines != value["lines"]:
                        _change = value["lines"] - old_lines
                        lines_change = "{:+}".format(_change)
                        lines_symbol = lines_change[0]
                        if _change > 0:
                            additions += _change
                        else:
                            deletions -= _change
                    else:
                        lines_symbol = lines_change = ""

                else:
                    files_change = files_symbol = lines_change = lines_symbol = ""

                file_symbol = self.symbol_color.get(files_symbol, "")
                line_symbol = self.symbol_color.get(lines_symbol, "")
                tb_data.append(
                    [
                        f" {key_display_str:<21}",
                        f" {Fx.i}{value['files']:<11}{Fx.rs} {file_symbol}{files_change:>5}{Fx.rs}",
                        f" {Fx.i}{value['lines']:<15}{Fx.rs} {line_symbol}{lines_change:>6}{Fx.rs}",
                    ]
                )

                # Clac sum code line.
                sum_ += value["lines"]

            # Print table.
            tb = Table(header, tb_data, title=title)
            tb.print()

            # Print total and change graph.
            print(" Total: {0}".format(sum_))

            # Additions and deletions are calculated by percentage,
            # and the base is the total number of last statistics.
            if additions > 0 or deletions > 0:
                # Get prev count sum.
                old_sum = sum([i["lines"] for i in old.values()])

                print(" Altered: ", end="")
                print(
                    "{0}{1}".format(
                        self.symbol_color["+"], "+" * ceil(additions / old_sum * 100)
                    ),
                    end="",
                )
                print(
                    "{0}{1}{2}".format(
                        self.symbol_color["-"],
                        "-" * ceil(deletions / old_sum * 100),
                        Fx.reset,
                    )
                )

    def count_and_format_print(self, if_save=True, show_invalid=False) -> None:
        result, invalid_list, total_size = self.count(self.count_path, self.use_ignore)

        old_result = self.load_recorded_result(self.count_path)

        # diff print.
        self.format_print(result, old_result)
        if if_save:
            self.save_result(result, self.count_path)
        if (
            show_invalid
            and invalid_list
            and confirm("Whether print invalid file list?[y/n]", default=False)
        ):
            print(invalid_list)

        # optimize size unit.
        size_unit = ["byte", "KB", "MB", "GB"]
        for i in range(3):
            if total_size >= 1024:
                total_size /= 1024
            else:
                break
        else:
            i = 3
        print(" Files total size: {0:.2f}{1}".format(total_size, size_unit[i]))



# -*- coding:utf-8 -*-

import sys
from typing import Tuple


class Color(object):
    """Holds representations for a 24-bit color value

    __init__(color, depth="fg", default=False)
        -- color accepts 6 digit hexadecimal: string "#RRGGBB", 2 digit
            hexadecimal: string "#FF" or decimal RGB "255 255 255" as a string.
        -- depth accepts "fg" or "bg"
    __call__(*args) joins str arguments to a string and apply color
    __str__ returns escape sequence to set color
    __iter__ returns iteration over red, green and blue in integer values of 0-255.

    * Values:
        .hexa: str
        .dec: Tuple[int, int, int]
        .red: int
        .green: int
        .blue: int
        .depth: str
        .escape: str
    """

    hexa: str
    dec: Tuple[int, int, int]
    red: int
    green: int
    blue: int
    depth: str
    escape: str
    default: bool

    TRUE_COLOR = sys.version_info < (3, 0)

    def __init__(self, color, depth="fg", default=False):
        self.depth = depth
        self.default = default
        try:
            if not color:
                self.dec = (-1, -1, -1)
                self.hexa = ""
                self.red = self.green = self.blue = -1
                self.escape = "\033[49m" if depth == "bg" and default else ""
                return

            elif color.startswith("#"):
                self.hexa = color
                if len(self.hexa) == 3:
                    self.hexa += self.hexa[1:3] + self.hexa[1:3]
                    c = int(self.hexa[1:3], base=16)
                    self.dec = (c, c, c)
                elif len(self.hexa) == 7:
                    self.dec = (
                        int(self.hexa[1:3], base=16),
                        int(self.hexa[3:5], base=16),
                        int(self.hexa[5:7], base=16),
                    )
                else:
                    raise ValueError(
                        "Incorrectly formatted hexadecimal rgb string: {}".format(
                            self.hexa
                        )
                    )

            else:
                c_t = tuple(map(int, color.split(" ")))
                if len(c_t) == 3:
                    self.dec = c_t  # type: ignore
                else:
                    raise ValueError('RGB dec should be "0-255 0-255 0-255"')

            ct = self.dec[0] + self.dec[1] + self.dec[2]
            if ct > 255 * 3 or ct < 0:
                raise ValueError("RGB values out of range: {}".format(color))
        except Exception:
            # errlog.exception(str(e))
            self.escape = ""
            return

        if self.dec and not self.hexa:
            self.hexa = "%s%s%s" % (
                hex(self.dec[0]).lstrip("0x").zfill(2),
                hex(self.dec[1]).lstrip("0x").zfill(2),
                hex(self.dec[2]).lstrip("0x").zfill(2),
            )

        if self.dec and self.hexa:
            self.red, self.green, self.blue = self.dec
            self.escape = "\033[%s;2;%sm" % (
                38 if self.depth == "fg" else 48,
                ";".join(str(c) for c in self.dec),
            )

        if Color.TRUE_COLOR:
            self.escape = "{}".format(
                self.truecolor_to_256(rgb=self.dec, depth=self.depth)
            )

    def __str__(self):
        return self.escape

    def __repr__(self):
        return repr(self.escape)

    def __iter__(self):
        for c in self.dec:
            yield c

    # def __call__(self, *args: str) -> str:
    #     if len(args) < 1:
    #         return ""
    #     return f'{self.escape}{"".join(args)}{getattr(Term, self.depth)}'

    @staticmethod
    def truecolor_to_256(rgb, depth="fg"):
        out = ""
        pre = "\033[{};5;".format("38" if depth == "fg" else "48")

        greyscale = (rgb[0] // 11, rgb[1] // 11, rgb[2] // 11)
        if greyscale[0] == greyscale[1] == greyscale[2]:
            out = "{}{}m".format(pre, 232 + greyscale[0])
        else:
            out = "{}{}m".format(
                pre,
                round(rgb[0] / 51) * 36
                + round(rgb[1] / 51) * 6
                + round(rgb[2] / 51)
                + 16,
            )

        return out

    @staticmethod
    def escape_color(hexa="", r=0, g=0, b=0, depth="fg"):
        """Returns escape sequence to set color
        * accepts either 6 digit hexadecimal hexa="#RRGGBB", 2 digit hexadecimal: hexa="#FF"
        * or decimal RGB: r=0-255, g=0-255, b=0-255
        * depth="fg" or "bg"
        """
        dint = 38 if depth == "fg" else 48
        color = ""
        if hexa:
            try:
                if len(hexa) == 3:
                    c = int(hexa[1:], base=16)
                    if Color.TRUE_COLOR:
                        color = "\033[{};2;{};{};{}m".format(dint, c, c, c)
                    else:
                        color = Color.truecolor_to_256(rgb=(c, c, c), depth=depth)
                elif len(hexa) == 7:
                    if Color.TRUE_COLOR:
                        color = "\033[{};2;{};{};{}m".format(
                            dint,
                            int(hexa[1:3], base=16),
                            int(hexa[3:5], base=16),
                            int(hexa[5:7], base=16),
                        )
                    else:
                        color = "{}".format(
                            Color.truecolor_to_256(
                                rgb=(
                                    int(hexa[1:3], base=16),
                                    int(hexa[3:5], base=16),
                                    int(hexa[5:7], base=16),
                                ),
                                depth=depth,
                            )
                        )
            except ValueError:
                # errlog.exception(f'{e}')
                pass
        else:
            if Color.TRUE_COLOR:
                color = "\033[{};2;{};{};{}m".format(dint, r, g, b)
            else:
                color = "{}".format(Color.truecolor_to_256(rgb=(r, g, b), depth=depth))
        return color

    @classmethod
    def fg(cls, *args):
        if len(args) > 2:
            return cls.escape_color(r=args[0], g=args[1], b=args[2], depth="fg")
        else:
            return cls.escape_color(hexa=args[0], depth="fg")

    @classmethod
    def bg(cls, *args):
        if len(args) > 2:
            return cls.escape_color(r=args[0], g=args[1], b=args[2], depth="bg")
        else:
            return cls.escape_color(hexa=args[0], depth="bg")


class Symbol(object):
    # yapf: disable
    rune:dict[str,list[str]] = {
        "normal":        ["-", "|", "|", "|", "|", "|", "|", "|", "-", "-", "-"],
        "normal_double": ["=", "‖", "‖", "‖", "‖", "‖", "‖", "‖", "=", "=", "="],
        "fine":          ["─", "│", "╭", "╮", "╰", "╯", "├", "┤", "┬", "┴", "┼"],
        "radian":        ["─", "│", "┌", "┐", "└", "┘", "├", "┤", "┬", "┴", "┼"],
        "bold":     ["━", "┃", "┏", "┓", "┗", "┛", "┣", "┫", "┳", "┻", "╋"],
        "double":   ["═", "║", "╔", "╗", "╚", "╝", "╠", "╣", "╦", "╩", "╬"],
    }

    normal_rune: list[str] = ["-", "|", "|", "|", "|", "|", "|", "|", "-", "-", "-"]
    normal_double_rune: list[str] = ["=", "‖", "‖", "‖", "‖", "‖", "‖", "‖", "=", "=", "="]
    fine_rune: list[str] = ["─", "│", "╭", "╮", "╰", "╯", "├", "┤", "┬", "┴", "┼"]
    radian_rune: list[str] = ["─", "│", "┌", "┐", "└", "┘", "├", "┤", "┬", "┴", "┼"]
    bold_rune: list[str] = ["━", "┃", "┏", "┓", "┗", "┛", "┣", "┫", "┳", "┻", "╋"]
    double_rune: list[str] = ["═", "║", "╔", "╗", "╚", "╝", "╠", "╣", "╦", "╩", "╬"]
    # yapf: enable



# -*- coding:utf-8 -*-
import os

# For windows print color.
if os.name == "nt":
    os.system("")

from .emotion import Emotion
from .style import Color, Symbol
from .escape import Fx, Cursor

__all__ = ["Emotion", "Color", "Symbol", "Fx", "Cursor"]


class TermColor:
    """Terminal print color class."""

    Red = Color.fg("#FF6347")  # Tomato
    Green = Color.fg("#98FB98")  # PaleGreen
    DeepGreen = Color.fg("#A4BE8C")  # PaleGreen
    Yellow = Color.fg("#EBCB8C")
    Gold = Color.fg("#FFD700")  # Gold
    SkyBlue = Color.fg("#87CEFA")
    MediumVioletRed = Color.fg("#C71585")
    Symbol = {"+": Color.fg("#98FB98"), "-": Color.fg("#FF6347")}



# -*- coding: utf-8 -*-

# Pomodoro 番茄工作法 https://en.wikipedia.org/wiki/Pomodoro_Technique
# ====== 🍅 Tomato Clock =======
# ./tomato.py         # start a 25 minutes tomato clock + 5 minutes break
# ./tomato.py -t      # start a 25 minutes tomato clock
# ./tomato.py -t <n>  # start a <n> minutes tomato clock
# ./tomato.py -b      # take a 5 minutes break
# ./tomato.py -b <n>  # take a <n> minutes break
# ./tomato.py -h      # help


import sys
import time
import subprocess

WORK_MINUTES = 25
BREAK_MINUTES = 5


def main(command_str=[]):
    if not command_str:
        args = sys.argv
    else:
        args = command_str
    try:
        if len(args) <= 1:
            print("🍅 tomato {} minutes. Ctrl+C to exit".format(WORK_MINUTES))
            tomato(WORK_MINUTES, "It is time to take a break")
            print("🛀 break {} minutes. Ctrl+C to exit".format(BREAK_MINUTES))
            tomato(BREAK_MINUTES, "It is time to work")

        elif args[1] == "-t":
            minutes = int(args[2]) if len(args) > 2 else WORK_MINUTES
            print("🍅 tomato {} minutes. Ctrl+C to exit".format(minutes))
            tomato(minutes, "It is time to take a break")

        elif args[1] == "-b":
            minutes = int(args[2]) if len(args) > 2 else BREAK_MINUTES
            print("🛀 break {} minutes. Ctrl+C to exit".format(minutes))
            tomato(minutes, "It is time to work")

        elif args[1] == "-h":
            help(args[0])

        else:
            help(args[0])

    except KeyboardInterrupt:
        print("\n👋 goodbye")
    except Exception as e:
        print(str(e), str(e.__traceback__))


def tomato(minutes, notify_msg):
    start_time = time.time()
    while True:
        diff_seconds = int(round(time.time() - start_time))
        left_seconds = minutes * 60 - diff_seconds
        if left_seconds <= 0:
            print("")
            break

        countdown = "{}:{} ⏰".format(int(left_seconds / 60), int(left_seconds % 60))
        duration = min(minutes, 25)
        progressbar(diff_seconds, minutes * 60, duration, countdown)
        time.sleep(1)

    notify_me(notify_msg)


def progressbar(current, total, duration=10, extra=""):
    frac = current / total
    filled = int(round(frac * duration))
    print(
        "\r",
        "🍅" * filled + "--" * (duration - filled),
        "[{:.0%}]".format(frac),
        extra,
        end="",
    )
    sys.stdout.flush()


def notify_me(msg):
    """
    # macos desktop notification
    terminal-notifier -> https://github.com/julienXX/terminal-notifier#download
    terminal-notifier -message <msg>
    # ubuntu desktop notification
    notify-send
    # voice notification
    say -v <lang> <msg>
    lang options:
    - Daniel:       British English
    - Ting-Ting:    Mandarin
    - Sin-ji:       Cantonese
    """

    print(msg)
    try:
        if sys.platform == "darwin":
            # macos desktop notification
            subprocess.run(["terminal-notifier", "-title", "🍅", "-message", msg])
            subprocess.run(["say", "-v", "Daniel", msg])
        elif sys.platform.startswith("linux"):
            # ubuntu desktop notification
            subprocess.Popen(["notify-send", "🍅", msg])
        else:
            # windows?
            # TODO: windows notification
            pass

    except:
        # skip the notification error
        pass


def help(appname):
    appname = appname if appname.endswith(".py") else "tomato"  # tomato is pypi package
    print("====== 🍅 Tomato Clock =======")
    print(
        "{0}         # start a {1} minutes tomato clock + {2} minutes break".format(
            appname, WORK_MINUTES, BREAK_MINUTES
        )
    )
    print(
        "{0} -t      # start a {1} minutes tomato clock".format(appname, WORK_MINUTES)
    )
    print("{0} -t <n>  # start a <n> minutes tomato clock".format(appname))
    print("{0} -b      # take a {1} minutes break".format(appname, BREAK_MINUTES))
    print("{0} -b <n>  # take a <n> minutes break".format(appname))
    print("{0} -h      # help".format(appname))


if __name__ == "__main__":
    main()



# -*- coding:utf-8 -*-

"""Command methods.

This file encapsulates some methods corresponding to command.
All methods are must and only accept an `args` parameter -- a list or tuple
of parameters to be processed. If don't need will get a empty list or tuple.
"""

import re
from typing import Union

from ..utils import exec_cmd, run_cmd, color_print
from ..common import Emotion, TermColor


def add(args: Union[list, tuple]) -> None:
    """git add option.

    Args:
        args (list): arguments list, maybe empty. need file string.
    """

    # default add all.
    args_str = " ."
    # process arguments if has.
    if args:
        args_str = " ".join(args)

    print(
        "{0} Storage file: {1}".format(
            Emotion.Icon_Rainbow, "all" if args_str.strip() == "." else args_str
        )
    )
    run_cmd("git add " + args_str)


def fetch_remote_branch(args: Union[list, tuple]):
    branch = args[0] if len(args) > 1 else None

    if branch:
        run_cmd("git fetch origin {0}:{0} ".format(branch))
    else:
        color_print("This option need a branch name.", TermColor.Red)


def set_email_and_username(args: Union[list, tuple]):
    """Set git username and email with interaction.

    Args:
        args (list): arguments list, maybe empty.
    """

    print("Set the interactive environment of user name and email ...")
    _global = re.compile(r"\-\-global")
    for i in args:
        r = _global.search(i)
        if r is not None:
            other = " --global "
            print("Now set for global.")
            break
    else:
        print("Now set for local.")
        other = " "

    name = input("Please input username:").strip()
    while True:
        if not name:
            color_print("Name is empty.", TermColor.Red)
            name = input("Please input username again:")
        else:
            break

    email = input("Please input email:")
    email_re = re.compile(r"^[0-9a-zA-Z_]{0,19}@[0-9a-zA-Z]{1,13}\.[com,cn,net]{1,3}$")
    while True:
        if email_re.match(email) is None:
            color_print("Bad mailbox format.", TermColor.Red)
            email = input("Please input email again:")
        else:
            break

    if run_cmd("git config user.name" + other + name) and run_cmd(
        "git config user.email" + other + email
    ):
        color_print("Successfully set.", TermColor.Green)
    else:
        color_print("Failed. Please check log.", TermColor.Red)


def open_remote_url(args: Union[list, tuple]) -> None:
    """Open remote.

    Open the remote repository through the browser. Support additional parameter list.
    When multiple parameters exist at the same time, only one will take effect.

    -i, --issue:
        open given issue of the repository.
        pigit open -- -i 20
        pigit open -- --issue=20
    -c, --commit:
        open the current commit in the repo website.
        pigit open -- --commit
    -p, --print:
        only print the url at the terminal, but don't open it.
    <branch>:
        open the page for this branch on the repo website.
    """

    err, branches = exec_cmd("git branch")

    branch = issue = commit = ""
    if args:
        i = 0
        while i < len(args):
            # -i 29 or --issue 29
            if args[i].strip() in ["-i", "--issue"]:
                try:
                    issue_number = args[i + 1]
                    int(issue_number)
                    issue = "/issues/{0}".format(issue_number)
                    i += 2
                except:
                    i += 1

            # -i=29 or --issue=29
            elif ("-i" in args[i] or "--issue" in args[i]) and "=" in args[i]:
                issue_number = args[i].split("=")[-1]
                try:
                    int(issue_number)
                    issue = "/issues/{0}".format(issue_number)
                finally:
                    i += 1
            elif args[i].strip() in ["-c", "--commit"]:
                err, commit_hash = exec_cmd("git log -n1 --format=format:'%H'")
                if len(commit_hash.strip()) == 40:
                    commit = "/commit/{0}".format(commit_hash.strip())
                i += 1
            else:
                if re.search(args[i], branches):
                    branch = "/tree/{0}".format(args[i])
                i += 1

    # Get remote name, exit when error.
    err, remote = exec_cmd("git remote show")
    if err:
        print(err)
        return None
    remote = remote.strip().split("\n")[0]

    # Get remote url, exit when error.
    err, remote_url = exec_cmd("git ls-remote --get-url {0}".format(remote))
    if err:
        print(err)
        return None
    remote_url = remote_url[:-5]

    # Splice URL, priority: branch > commit > issue
    if branch:
        remote_url += branch
    elif commit:
        remote_url += commit
    elif issue:
        remote_url += issue

    # Adjust whether just need print.
    if "-p" in args or "--print" in args:
        print(remote_url)
    else:
        import webbrowser

        webbrowser.open(remote_url + branch)



# -*- coding:utf-8 -*-

"""This file support some str util method."""

# yapf: disable
WIDTHS:list[tuple[int,int]] = [
    (126, 1), (159, 0), (687, 1), (710, 0), (711, 1), (727, 0), (733, 1), (879, 0),
    (1154, 1), (1161, 0), (4347, 1), (4447, 2), (7467, 1),
    (7521, 0), (8369, 1), (8426, 0), (9000, 1), (9002, 2),
    (11021, 1), (12350, 2), (12351, 1), (12438, 2), (12442, 0), (19893, 2),
    (19967, 1), (55203, 2), (63743, 1), (64106, 2), (65039, 1), (65059, 0),
    (65131, 2), (65279, 1), (65376, 2), (65500, 1), (65510, 2), (120831, 1),
    (262141, 2), (1114109, 1),
]
# yapf: enable


def get_width(r: int) -> int:
    """Gets the width occupied by characters on the command line.

    >>> get_width(ord('a'))
    1
    >>> get_width(ord('中'))
    2
    >>> get_width(ord('Ç'))
    1
    """
    global WIDTHS
    if r == 0xE or r == 0xF:
        return 0
    for num, wid in WIDTHS:
        if r <= num:
            return wid
    return 1


# Mark the type corresponding to the file suffix.
# abcdefg hijklmn opq rst uvw xyz
SUFFIX_TYPE: dict[str, str] = {
    "": "",
    "bat": "Batch",
    "c": "C",
    "cfg": "Properties",
    "conf": "Properties",
    "cpp": "C++",
    "cs": "C#",
    "css": "CSS",
    "dart": "Dart",
    "dea": "XML",
    "go": "Go",
    "gradle": "Groovy",
    "h": "C",
    "hpp": "C++",
    "htm": "HTML",
    "html": "HTML",
    "java": "Java",
    "js": "Java Script",
    "json": "Json",
    "jsx": "React",
    "kt": "Kotlin",
    "launch": "XML",
    "less": "CSS",
    "lua": "Lua",
    "markdown": "Markdown",
    "md": "Markdown",
    "msg": "ROS Message",
    "php": "PHP",
    "plist": "XML",
    "properties": "Propertie",
    "py": "Python",
    "r": "R",
    "rb": "Ruby",
    "rc": "Properties",
    "rs": "Rust",
    "rst": "reStructuredText",
    "rviz": "YAML",
    "ts": "Type Script",
    "tsx": "React",
    "sass": "CSS",
    "scss": "CSS",
    "sh": "Shell",
    "sql": "SQL",
    "srdf": "YAML",
    "srv": "ROS Message",
    "swift": "Swift",
    "toml": "Properties",
    "urdf": "XML",
    "vim": "Vim Scirpt",
    "vue": "Vue",
    "xhtml": "HTML",
    "xml": "XML",
    "yaml": "YAML",
    "yml": "YAML",
    "zsh": "Shell",
}

# Mark the type of some special files.
SPECIAL_NAMES: dict[str, str] = {
    "license": "LICENSE",
    "requirements.txt": "Pip requirement",
    "vimrc": "Vim Scirpt",
}


def adjudgment_type(file: str, original: bool = False) -> str:
    """Get file type.

    First, judge whether the file name is special, and then query the
    file suffix. Otherwise, the suffix or name will be returned as is.

    Args:
        file (str): file name string.
        original (bool, option): whether return origin when match failed.

    Returns:
        (str): file type.

    >>> adjudgment_type('test.py')
    'Python'
    >>> adjudgment_type('xxxxx')
    'unknown'
    >>> adjudgment_type('xxxxx', True)
    'xxxxx'
    """

    pre_type = SPECIAL_NAMES.get(file.lower(), None)
    if pre_type:
        return pre_type

    suffix = file.split(".")[-1]
    suffix_type = SUFFIX_TYPE.get(suffix.lower(), None)
    if suffix_type:
        return suffix_type
    else:
        return suffix if original else "unknown"


FILE_ICONS: dict[str, str] = {
    "": "",
    "Batch": "",
    "C": "",
    "C#": "",
    "C++": "",
    "CSS": "",
    "Dart": "",
    "Groovy": "",
    "Go": "",
    "HTML": "",
    "Java": "",
    "Java Scirpt": "",
    "Lua": "",
    "Kotlin": "",
    "Markdown": "",
    "PHP": "",
    "Propertie": "",
    "Python": "",
    "R": "ﳒ",
    "React": "",
    "Ruby": "",
    "Rust": "",
    "ROS Message": "",
    "reStructuredText": "",
    "Shell": "",
    "Swift": "",
    "SQL": "",
    "snippets": "",
    "Type Scirpt": "",
    "Vim Scirpt": "",
    "Vue": "﵂",
    "YAML": "",
    "XML": "",
}


def get_file_icon(file_type: str) -> str:
    """According file type return icon.

    Args:
        file_type (str): type string.

    Returns:
        str: icon.

    >>> get_file_icon('Python')
    ''
    >>> get_file_icon('Lua')
    ''
    >>> get_file_icon('xxxxxxxxxx')
    ''
    """

    #     
    return FILE_ICONS.get(file_type, "")


def shorten(
    text: str, width: int, placeholder: str = "...", front: bool = False
) -> str:
    """Truncate exceeded characters.

    Args:
        text (str): Target string.
        width (int): Limit length.
        placeholder (str): Placeholder string. Defaults to "..."
        front (bool): Head hidden or tail hidden. Defaults to False.

    Returns:
        (str): shorten string.

    >>> shorten('Hello world!', 9, placeholder='^-^')
    'Hello ^-^'
    >>> shorten('Hello world!', 9, placeholder='^-^', front=True)
    '^-^world!'
    """

    if len(text) > width:
        if front:
            text = placeholder + text[-width + len(placeholder) :]
        else:
            text = text[: width - len(placeholder)] + placeholder

    return text


if __name__ == "__main__":
    import doctest

    doctest.testmod()



# -*- coding:utf-8 -*-

import os
import time
from math import ceil
from shutil import get_terminal_size
from typing import Optional

from ..utils import color_print, exec_cmd, run_cmd, confirm
from ..common import Fx, Color, TermColor
from ..common.str_utils import shorten, get_width
from ..keyevent import get_keyevent_obj
from ..git_utils import REPOSITORY_PATH
from .model import File


#####################################################################
# Part of command.                                                  #
#####################################################################
class TermError(Exception):
    pass


class DataHandle(object):
    def __init__(self, use_color: bool):
        self.use_color = use_color

    def get_status(self, max_width: int, ident: int = 2) -> list[File]:
        """Get the file tree status of GIT for processing and encapsulation.

        Args:
            max_width (int): The max length of display string.
            ident (int, option): Number of reserved blank characters in the header.

        Raises:
            Exception: Can't get tree status.

        Returns:
            (list[File]): Processed file status list.
        """

        file_items = []
        err, files = exec_cmd("git status -s -u --porcelain")
        if err:
            raise Exception("Can't get git status.")
        for file in files.rstrip().split("\n"):
            if not file.strip():
                # skip blank line.
                continue
            change = file[:2]
            staged_change = file[:1]
            unstaged_change = file[1:2]
            name = file[3:]
            untracked = change in ["??", "A ", "AM"]
            has_no_staged_change = staged_change in [" ", "U", "?"]
            has_merged_conflicts = change in ["DD", "AA", "UU", "AU", "UA", "UD", "DU"]
            has_inline_merged_conflicts = change in ["UU", "AA"]

            display_name = shorten(name, max_width - 3 - ident)
            # color full command.
            if unstaged_change != " ":
                if not has_no_staged_change:
                    display_str = "{}{}{}{} {}{}".format(
                        TermColor.Green,
                        staged_change,
                        TermColor.Red,
                        unstaged_change,
                        display_name,
                        Fx.reset,
                    )
                else:
                    display_str = "{}{} {}{}".format(
                        TermColor.Red, change, display_name, Fx.reset
                    )
            else:
                display_str = "{}{} {}{}".format(
                    TermColor.Green, change, display_name, Fx.reset
                )

            file_ = File(
                name=name,
                display_str=display_str if self.use_color else file,
                short_status=change,
                has_staged_change=not has_no_staged_change,
                has_unstaged_change=unstaged_change != " ",
                tracked=not untracked,
                deleted=unstaged_change == "D" or staged_change == "D",
                added=unstaged_change == "A" or untracked,
                has_merged_conflicts=has_merged_conflicts,
                has_inline_merged_conflicts=has_inline_merged_conflicts,
            )
            file_items.append(file_)

        return file_items

    def get_file_diff(
        self, file: str, tracked: bool = True, cached: bool = False, plain: bool = False
    ) -> str:
        """Gets the modification of the file.

        Args:
            file (str): file path relative to git.
            tracked (bool, optional): Defaults to True.
            cached (bool, optional): Defaults to False.
            plain (bool, optional): Whether need color. Defaults to False.

        Returns:
            (str): change string.
        """

        command = "git diff --submodule --no-ext-diff {plain} {cached} {tracked} {file}"

        if plain:
            _plain = "--color=never"
        else:
            _plain = "--color=always"

        if cached:
            _cached = "--cached"
        else:
            _cached = ""

        if not tracked:
            _tracked = "--no-index -- /dev/null"
        else:
            _tracked = "--"

        if "->" in file:  # rename status.
            file = file.split("->")[-1].strip()

        err, res = exec_cmd(
            command.format(plain=_plain, cached=_cached, tracked=_tracked, file=file)
        )
        if err:
            return "Can't get diff."
        return res.rstrip()


class InteractiveAdd(object):
    """Interactive operation git tree status."""

    def __init__(
        self,
        use_color: bool = True,
        cursor: Optional[str] = None,
        help_wait: float = 1.5,
        debug: bool = False,
    ) -> None:
        super(InteractiveAdd, self).__init__()
        self.use_color = use_color

        if not cursor:
            self.cursor = "→"
        elif len(cursor) == 1:
            if get_width(ord(cursor)) == 1:
                self.cursor = cursor
            else:
                self.cursor = "→"
                print("When displayed, it will occupy more than one character.")
        else:
            self.cursor = "→"
            print("The cursor symbol entered is not supported.")

        self.help_wait = help_wait
        self._min_height = 8
        self._min_width = 60
        self._debug = debug

        # Whether can into interactive.
        try:
            _keyevent_class = get_keyevent_obj()
        except NameError:
            raise TermError("This behavior is not supported in the current system.")
        self._keyevent = _keyevent_class()

        self._data_handle = DataHandle(use_color)

    def process_file(self, file: File) -> None:
        """Process file to change the status.

        Args:
            file (File): One processed file.
        """

        if file.has_merged_conflicts or file.has_inline_merged_conflicts:
            pass
        elif file.has_unstaged_change:
            exec_cmd("git add -- {}".format(file.name))
        elif file.has_staged_change:
            if file.tracked:
                exec_cmd("git reset HEAD -- {}".format(file.name))
            else:
                exec_cmd("git rm --cached --force -- {}".format(file.name))

    def show_diff(self, file_obj: File) -> None:
        """Interactive display file diff.

        Args:
            file_obj (File): needed file.

        Raises:
            TermError: terminal size not enough.
        """

        width, height = get_terminal_size()

        # Initialize.
        cursor_row = 1
        display_range = [1, height - 1]

        stopping = False  # exit signal.

        def _process_diff(diff_list, term_width):
            """Process diff raw list.

            Generate a new list, in which each element is a tuple in the shape of (str, int).
            The first parameter is the displayed string, and the second is the additional
            row to be occupied under the current width.
            """
            new_list = []
            for line in diff_list:
                text = Fx.uncolor(line)
                count = 0
                for ch in text:
                    count += get_width(ord(ch))
                # [float] is to solve the division of python2 without retaining decimal places.
                new_list.append((line, ceil(count / term_width) - 1))
            return new_list

        # only need get once.
        diff_raw = self._data_handle.get_file_diff(
            file_obj.name,
            file_obj.tracked,
            file_obj.has_staged_change,
            not self.use_color,
        ).split("\n")

        diff_ = _process_diff(diff_raw, width)
        if self._debug:  # debug mode print all occupied line num.
            print(Fx.clear_)
            print(str([i[1] for i in diff_]))
            input()

        extra = 0  # Extra occupied row.
        while not stopping:
            print(Fx.clear_)

            while cursor_row < display_range[0]:
                display_range = [i - 1 for i in display_range]
            while cursor_row + extra > display_range[1]:
                display_range = [i + 1 for i in display_range]

            extra = 0  # Return to zero and accumulate again.
            # Terminal outputs the text to be displayed.
            for index, data in enumerate(diff_, start=1):
                line, each_extra = data
                if display_range[0] <= index <= display_range[1] - extra:
                    if index == cursor_row:
                        print("{}{}{}".format(Color.bg("#6495ED"), line, Fx.reset))
                    else:
                        print(line)
                    extra += each_extra

            input_key = self._keyevent.sync_get_input()
            if input_key in ["q", "escape"]:
                # exit.
                stopping = True
            elif input_key in ["j", "down"]:
                # select pre file.
                cursor_row += 1
                cursor_row = cursor_row if cursor_row < len(diff_) else len(diff_)
            elif input_key in ["k", "up"]:
                # select next file.
                cursor_row -= 1
                cursor_row = cursor_row if cursor_row > 1 else 1
            elif input_key in ["J"]:
                # scroll down 5 lines.
                cursor_row += 5
                cursor_row = cursor_row if cursor_row < len(diff_) else len(diff_)
            elif input_key in ["K"]:
                # scroll up 5 line
                cursor_row -= 5
                cursor_row = cursor_row if cursor_row > 1 else 1
            elif input_key == "windows resize":
                # get new term height.
                new_width, new_height = get_terminal_size()
                if new_height < self._min_height or new_width < self._min_width:
                    raise TermError(
                        "The minimum size of terminal should be {0} x {1}.".format(
                            self._min_width, self._min_height
                        )
                    )
                # get size diff, reassign.
                line_diff = new_height - height
                width, height = new_width, new_height
                # get new display range.
                display_range[1] += line_diff
                diff_ = _process_diff(diff_raw, width)
            elif input_key in ["?", "h"]:
                # show help messages.
                print(Fx.clear_)
                print(
                    (
                        "k / ↑: select previous line.\n"
                        "j / ↓: select next line.\n"
                        "J: Scroll down 5 lines.\n"
                        "K: Scroll down 5 lines.\n"
                        "? : show help, wait {}s and exit.\n"
                    ).format(self.help_wait)
                )
                if self.help_wait == 0:
                    self._keyevent.sync_get_input()
                else:
                    time.sleep(self.help_wait)
            else:
                continue

    def discard_changed(self, file: File) -> None:
        """Discard file all changed.

        Args:
            file (File): file object.
        """
        print(Fx.clear_)
        if confirm("discard all changed? [y/n]:"):
            if file.tracked:
                exec_cmd("git checkout -- {}".format(file.name))
            else:
                os.remove(os.path.join(file.name))

    def add_interactive(self, *args):
        """Interactive main method."""

        if not REPOSITORY_PATH:
            color_print("Current path is not a git repository.", TermColor.Red)
            return

        width, height = get_terminal_size()
        if height < self._min_height or width < self._min_width:
            raise TermError(
                "The minimum size of terminal should be {0} x {1}.".format(
                    self._min_width, self._min_height
                )
            )

        if self._debug:  # debug show.
            print(Fx.clear_)
            print(width, height)
            time.sleep(1.5)

        # Initialize.
        cursor_row: int = 1
        cursor_icon: str = self.cursor
        display_range: list = [1, height - 1]

        stopping: bool = False

        file_items = self._data_handle.get_status(width)
        if not file_items:
            print("The work tree is clean and there is nothing to operate.")
            return

        # Into new term page.
        print(Fx.alt_screen + Fx.hide_cursor)
        try:
            #  try hook window resize event.
            self._keyevent.signal_init()

            # Start interactive.
            while not stopping:
                print(Fx.clear_)

                # check whether have status.
                if not file_items:
                    print("The work tree is clean and there is nothing to operate.")
                    time.sleep(1)
                    return

                while cursor_row < display_range[0]:
                    display_range = [i - 1 for i in display_range]
                while cursor_row > display_range[1]:
                    display_range = [i + 1 for i in display_range]

                # Print needed display part.
                for index, file in enumerate(file_items, start=1):
                    if display_range[0] <= index <= display_range[1]:
                        if index == cursor_row:
                            print("{} {}".format(cursor_icon, file.display_str))
                        else:
                            print("  " + file.display_str)

                input_key = self._keyevent.sync_get_input()
                if input_key in ["q", "escape"]:
                    # exit.
                    stopping = True
                elif input_key in ["j", "down"]:
                    # select pre file.
                    cursor_row += 1
                    cursor_row = (
                        cursor_row if cursor_row < len(file_items) else len(file_items)
                    )
                elif input_key in ["k", "up"]:
                    # select next file.
                    cursor_row -= 1
                    cursor_row = cursor_row if cursor_row > 1 else 1
                elif input_key in ["a", "space"]:
                    self.process_file(file_items[cursor_row - 1])
                    file_items = self._data_handle.get_status(width)
                elif input_key == "d":
                    self.discard_changed(file_items[cursor_row - 1])
                    file_items = self._data_handle.get_status(width)
                elif input_key == "e":
                    editor = os.environ.get("EDITOR", None)
                    if editor:
                        run_cmd(
                            '{} "{}"'.format(editor, file_items[cursor_row - 1].name)
                        )
                        file_items = self._data_handle.get_status(width)
                    else:
                        pass
                elif input_key == "enter":
                    self.show_diff(file_items[cursor_row - 1])
                elif input_key == "windows resize":
                    # get new term height.
                    new_width, new_height = get_terminal_size()
                    if new_height < self._min_height or new_width < self._min_width:
                        raise TermError(
                            "The minimum size of terminal should be {0} x {1}.".format(
                                self._min_width, self._min_height
                            )
                        )
                    # get diff, reassign.
                    line_diff = new_height - height
                    height = new_height
                    # get new display range.
                    display_range[1] += line_diff
                elif input_key == "?":
                    print(Fx.clear_)
                    print(
                        (
                            "k / ↑: select previous file.\n"
                            "j / ↓: select next file.\n"
                            "a / space: toggle storage or unstorage file.\n"
                            "d: discard the file changed.\n"
                            "e: open file with default editor.\n"
                            "↲ : check file diff.\n"
                            "? : show help, wait {}s and exit.\n"
                        ).format(self.help_wait)
                    )
                    if self.help_wait == 0:
                        self._keyevent.sync_get_input()
                    else:
                        time.sleep(self.help_wait)
                else:
                    # If not needed input key, skip and wait.
                    continue
        except KeyboardInterrupt:
            pass
        finally:
            # Whatever, unregister signal event and restore terminal at last.
            self._keyevent.signal_restore()
            print(Fx.normal_screen + Fx.show_cursor)



# -*- coding:utf-8 -*-

import os
import re
import signal
import subprocess
import logging
from math import sqrt
from collections import Counter
from typing import Union, Iterable


Log = logging.getLogger(__name__)

# Exit code.
EXIT_NORMAL = 0
EXIT_ERROR = 1


def leave(code: int, *args) -> None:
    """Exit program.

    Receive error code, error message. If the error code matches, print the
    error information to the log. Then the command line output prompt, and
    finally exit.

    Args:
        code: Exit code.
        *args: Other messages.
    """

    if code == EXIT_ERROR:
        Log.error(args)

    raise SystemExit(0)


def run_cmd(*args) -> bool:
    """Run system command.

    Returns:
        (bool): Whether run successful.

    >>> run_cmd('pwd')
    True
    >>> run_cmd('which', 'python')
    True
    """

    try:
        # ? In python2, `subprocess` not support `with` sentence.
        proc = subprocess.Popen(" ".join(args), shell=True)
        proc.wait()
    except Exception as e:
        Log.error(str(e) + str(e.__traceback__))
        return False
    else:
        return True


def exec_cmd(*args) -> tuple[str, str]:
    """Run system command and get result.

    Returns:
        (str, str): Error string and result string.
    """

    try:
        # Take over the input stream and get the return information.
        proc = subprocess.Popen(
            " ".join(args),
            stderr=subprocess.PIPE,
            stdout=subprocess.PIPE,
            shell=True,
        )

        # Get normal output and error output.
        res = proc.stdout.read().decode()
        err = proc.stderr.read().decode()
        proc.kill()
    except Exception as e:
        Log.error(str(e) + str(e.__traceback__))
        return str(e), ""
    else:
        return err, res


def confirm(text: str = "", default: bool = True) -> bool:
    """Obtain confirmation results.
    Args:
        text (str): Confirmation prompt.
        default (bool): Result returned when unexpected input.

    Returns:
        (bool): Confirm result.

    >>> confirm()
    True
    >>> confirm(default=False)
    False
    """
    input_command: str = input(text).strip().lower()
    if input_command in ["n", "no", "N", "No"]:
        return False
    elif input_command in ["y", "yes", "Y", "Yes"]:
        return True
    else:
        return default


def similar_command(command: str, all_commands: Iterable) -> str:
    """Get the most similar command with K-NearestNeighbor.

    Args:
        command (str): command string.
        all_commands (list): The list of all command.

    Returns:
        (str): most similar command string.

    >>> commands = ['branch', 'working tree', 'index', 'log', 'push', 'pull', 'tag','commit','conflict']
    >>> similar_command('br', commands)
    'branch'
    >>> similar_command('wo', commands)
    'working tree'
    >>> similar_command('com', commands)
    'commit'
    """

    #  The dictionary of letter frequency of all commands.
    words: dict = {word: dict(Counter(word)) for word in all_commands}
    # Letter frequency of command.
    fre = dict(Counter(command))
    # The distance between the frequency of each letter in the command
    # to be tested and all candidate commands, that is the difference
    # between the frequency of letters.
    frequency_difference = {
        word: [fre[ch] - words[word].get(ch, 0) for ch in command]
        + [words[word][ch] - fre.get(ch, 0) for ch in word]
        for word in words
    }
    # Square of sum of squares of word frequency difference.
    frequency_sum_square = list(
        map(
            lambda item: [item[0], sqrt(sum(map(lambda i: i ** 2, item[1])))],
            frequency_difference.items(),
        )
    )

    # The value of `frequency_sum_square` is multiplied by the weight to find
    # the minimum.
    # Distance weight: compensate for the effect of length difference.
    # Compare Weight: The more similar the beginning, the higher the weight.
    min_frequency_command: str = min(
        frequency_sum_square,
        key=lambda item: item[1]
        * (
            len(command) / len(item[0])
            if len(command) / len(item[0])
            else len(item[0]) / len(command)
        )
        # Returns how many identical letters are compared from the head. sigmod to 0 ~ 1.
        * (1 / (len(list(filter(lambda i: i[0] == i[1], zip(command, item[0])))) + 1)),
    )[0]
    return min_frequency_command


def color_print(value: str, *styles, **options) -> None:
    """Print to terminal.

    Print special information with color and style according to the
    incoming parameters.

    Args:
        msg: A special message.
        style: Message style, like: [bold, underline].
    """

    value = "{0}{1}\033[0m".format("".join(styles), value)
    end = options.get("end", "\n")
    print(value, end=end)


Color_Re = re.compile(r"^#[0-9A-Za-z]{6}")


def is_color(s: Union[str, list, tuple, None]) -> bool:
    """Adjust `s` whether is color. Like: '#FF0000', [255, 0, 0], (0, 255, 0)

    >>> is_color('#FF0000')
    True
    >>> is_color([255, 0, 0])
    True
    >>> is_color((0, 255, 0))
    True
    >>> is_color(None)
    False
    >>> is_color(12345)
    False
    """
    if not s:
        return False
    elif type(s) == str:
        return True if Color_Re.match(s) else False
    elif isinstance(s, list) or isinstance(s, tuple):
        if len(s) != 3:
            return False
        else:
            for i in s:
                if i < 0 or i > 255:
                    return False
            else:
                return True
    else:
        return False


def dir_whether_ok(dir_path: str) -> bool:
    """Determine whether the dir path exists. If not, create a directory.

    Args:
        dir_path (str): Directory path, like: "~/.config/xxx"

    >>> dir_whether_ok('.')
    True
    """

    if os.path.isdir(dir_path):
        return True
    try:
        os.makedirs(dir_path, exist_ok=True)
    except Exception as e:
        Log.error(str(e) + str(e.__traceback__))
        return False
    else:
        return True


def init_hook():
    """Take over some system signals, which are used at
    the beginning of PIGIT.
    """
    try:
        signal.signal(signal.SIGINT, leave)
    except Exception as e:
        print(str(e))


if __name__ == "__main__":
    import doctest

    doctest.testmod()



# -*- coding:utf-8 -*-

import os
import sys
import signal
from typing import Optional, Type

# try to import needed pkg.
# If import failed, set `TERM_CONTROL` is False.
TERM_CONTROL = True
NEED_EXTRA_KEYBOARD_EVENT_PKG = False
try:
    import select, termios, fcntl, tty  # noqa: E401

    B = lambda x: x.encode("iso8859-1")  # noqa: E731
except Exception:
    try:
        from pynput import keyboard  # type: ignore

        import queue

        NEED_EXTRA_KEYBOARD_EVENT_PKG = True
    except ModuleNotFoundError:
        TERM_CONTROL = False
        print("Please install `pynput` to supported keyboard event.")


class _baseKeyEvent(object):
    def signal_init(self) -> ...:
        """Take over the system signal, which is implemented by subclasses."""
        raise NotImplementedError()

    def signal_restore(self) -> ...:
        """Reply signal, implemented by subclass."""
        raise NotImplementedError()

    def sync_get_input(self) -> ...:
        raise NotImplementedError()


#####################################################################
# KeyBoard event classes.                                           #
#####################################################################
class _PosixKeyEvent(_baseKeyEvent):
    """KeyBoard event class.

    Subclass:
        Raw: Set raw input mode for device.
        Nonblocking: Set nonblocking mode for device.

    Attributes:
        escape: Translation dictionary.
        _resized: windows resize handle.

    Functions:
        signal_init: Register signal events.
        signal_restore: Unregister signal events.
        sync_get_key: get one input value, will wait until get input.
    """

    class Raw(object):
        """Set raw input mode for device"""

        def __init__(self, stream):
            self.stream = stream
            self.fd = self.stream.fileno()

        def __enter__(self):
            # Get original fd descriptor.
            self.original_descriptor = termios.tcgetattr(self.stream)
            # Change the mode of the file descriptor fd to cbreak.
            tty.setcbreak(self.stream)

        def __exit__(self, type, value, traceback):
            termios.tcsetattr(self.stream, termios.TCSANOW, self.original_descriptor)

    class Nonblocking(object):
        """Set nonblocking mode for device"""

        def __init__(self, stream):
            self.stream = stream
            self.fd = self.stream.fileno()

        def __enter__(self):
            self.orig_fl = fcntl.fcntl(self.fd, fcntl.F_GETFL)
            fcntl.fcntl(self.fd, fcntl.F_SETFL, self.orig_fl | os.O_NONBLOCK)

        def __exit__(self, *args):
            fcntl.fcntl(self.fd, fcntl.F_SETFL, self.orig_fl)

    escape = {
        "\n": "enter",
        ("\x7f", "\x08"): "backspace",
        ("[A", "OA"): "up",
        ("[B", "OB"): "down",
        ("[D", "OD"): "left",
        ("[C", "OC"): "right",
        "[2~": "insert",
        "[3~": "delete",
        "[H": "home",
        "[F": "end",
        "[5~": "page_up",
        "[6~": "page_down",
        "\t": "tab",
        "[Z": "shift_tab",
        "OP": "f1",
        "OQ": "f2",
        "OR": "f3",
        "OS": "f4",
        "[15": "f5",
        "[17": "f6",
        "[18": "f7",
        "[19": "f8",
        "[20": "f9",
        "[21": "f10",
        "[23": "f11",
        "[24": "f12",
        " ": "space",
    }

    _resize_pipe_rd, _resize_pipe_wr = os.pipe()
    _resized = False

    @classmethod
    def _sigwinch_handler(cls, signum, frame=None):
        if not cls._resized:
            os.write(cls._resize_pipe_wr, B("R"))
        cls._resized = True

    @classmethod
    def signal_init(cls):
        signal.signal(signal.SIGWINCH, cls._sigwinch_handler)

    @classmethod
    def signal_restore(cls):
        signal.signal(signal.SIGWINCH, signal.SIG_DFL)

    @classmethod
    def sync_get_input(cls):
        while True:
            with cls.Raw(sys.stdin):
                if cls._resized:
                    cls._resized = False
                    clean_key = "windows resize"
                    return clean_key

                # * Wait 100ms for input on stdin then restart loop to check for stop flag
                if not select.select([sys.stdin], [], [], 0.1)[0]:
                    continue
                input_key = sys.stdin.read(1)
                if input_key == "\033":
                    # * Set non blocking to prevent read stall
                    with cls.Nonblocking(sys.stdin):
                        input_key += sys.stdin.read(20)
                        if input_key.startswith("\033[<"):
                            _ = sys.stdin.read(1000)
                # print(repr(input_key))
                if input_key == "\033":
                    clean_key = "escape"
                elif input_key == "\\":
                    clean_key = "\\"  # * Clean up "\" to not return escaped
                else:
                    for code in cls.escape:
                        if input_key.lstrip("\033").startswith(code):
                            clean_key = cls.escape[code]
                            break
                    else:
                        clean_key = input_key

                # print(clean_key)
                return clean_key


class _WinKeyEvent(_baseKeyEvent):
    _special_keys = {
        "Key.space": "space",
        "Key.up": "up",
        "Key.down": "down",
        "Key.esc": "escape",
        "Key.enter": "enter",
    }

    def __init__(self):
        super(_WinKeyEvent, self).__init__()
        self._queue = queue.Queue(maxsize=1)
        self.listener = keyboard.Listener(on_press=self._on_press, suppress=True)
        self.listener.start()

    def _on_press(self, key):
        # process and push to queue

        self._queue.put(key)

    def signal_init(self) -> ...:
        pass

    def signal_restore(self) -> ...:
        pass

    def sync_get_input(self):
        resp = str(self._queue.get(block=True, timeout=None)).strip("'")
        if resp in self._special_keys:
            return self._special_keys[resp]
        else:
            return resp


def get_keyevent_obj() -> Type[_baseKeyEvent]:
    """Keyevent Object Factory."""
    if not TERM_CONTROL:
        raise NameError("Can't get right class.")
    if not NEED_EXTRA_KEYBOARD_EVENT_PKG:
        keyevent = _PosixKeyEvent
    else:
        keyevent = _WinKeyEvent
    return keyevent



# -*- coding:utf-8 -*-
import sys

# For encoding.
Icon_Supported_Encoding: list = ["utf-8"]


class Emotion(object):
    # XXX(zachary): There are some problems with the output emotion on windows.
    # ? In CMD, encoding is right, but emotion is error.
    # ? In git bash, encoding is not right, but seem can't detection.
    if (
        not sys.platform.lower().startswith("win")
        and sys.getdefaultencoding().lower() in Icon_Supported_Encoding
    ):
        Icon_Rainbow = "🌈"
        Icon_Smiler = "😊"
        Icon_Thinking = "🧐"
        Icon_Sorry = "😅"
    else:
        Icon_Rainbow = "::"
        Icon_Smiler = "^_^"
        Icon_Thinking = "-?-"
        Icon_Sorry = "Orz"



# -*- coding:utf-8 -*-

import time
from functools import wraps


def time_it(fn):
    """Print the overall running time.
    When recursive calls exist, only the outermost layer is printed.
    """
    time_it.deep = 0  # Mark recursion levels.
    time_unit = ["second", "mintue", "hour"]

    @wraps(fn)
    def wrap_(*args, **kwargs):
        time_it.deep += 1
        start_time = time.time()
        res = None
        try:
            res = fn(*args, **kwargs)
        except (SystemExit, EOFError):
            pass
        time_it.deep -= 1

        # Indicates that the decorated method does not or end a recursive call.
        if time_it.deep == 0:
            used_time = time.time() - start_time

            # Do unit optimization.
            for i in range(2):
                if used_time >= 60:
                    used_time /= 60
                else:
                    break
            else:
                i = 2
            print("\nruntime: {0:.2f} {1}".format(used_time, time_unit[i]))
        return res

    return wrap_



# -*- coding:utf-8 -*-

import random
from typing import Optional, Union

from ..utils import run_cmd, color_print, confirm, similar_command
from ..common import Fx, Color, TermColor, Emotion
from ..common.str_utils import shorten
from .interaction import InteractiveAdd, TermError
from .cmds import Git_Cmds, CommandType


class GitProcessor(object):
    """Git short command processor."""

    def __init__(
        self,
        extra_cmds: Optional[dict] = None,
        use_recommend: bool = True,
        show_original: bool = True,
        **kwargs
    ) -> None:
        super(GitProcessor, self).__init__()

        self.use_recommend = use_recommend
        self.show_original = show_original

        self.cmds = Git_Cmds
        self.cmds.update(
            {
                "i": {
                    "belong": CommandType.Index,
                    "command": InteractiveAdd(
                        use_color=kwargs.get("use_color", True),
                        help_wait=kwargs.get("help_wait", 1.5),
                    ).add_interactive,
                    "help": "interactive operation git tree status.",
                    "type": "func",
                },
                "shell": {
                    "command": "",
                    "help": "Into PIGIT shell mode.",
                },  # only for tips.
            }
        )

        if extra_cmds:
            if not isinstance(extra_cmds, dict):
                raise TypeError("Custom cmds must be dict.")
            self.cmds.update(extra_cmds)

    @staticmethod
    def color_command(command: str) -> str:
        """Color the command string.
        prog: green;
        short command: yellow;
        arguments: skyblue;
        values: white.

        Args:
            command(str): valid command string.

        Returns:
            (str): color command string.
        """

        command_list = command.split(" ")
        color_command = (
            Fx.bold
            + TermColor.DeepGreen
            + command_list.pop(0)
            + " "
            + TermColor.Yellow
            + command_list.pop(0)
            + " "
            + Fx.unbold
            # + Fx.italic
            # + TermColor.SkyBlue
        )

        less_len = len(command_list)
        idx = 0
        while idx < less_len:
            if command_list[idx].startswith("-"):
                color_command += "{}{}".format(Fx.italic, TermColor.SkyBlue)
                while idx < less_len and command_list[idx].startswith("-"):
                    color_command += command_list[idx] + " "
                    idx += 1
            else:
                color_command += Fx.reset
                while idx < less_len and not command_list[idx].startswith("-"):
                    color_command += command_list[idx] + " "
                    idx += 1

        color_command += Fx.reset

        return color_command

    def process_command(
        self, _command: str, args: Optional[Union[list, tuple]] = None
    ) -> None:
        """Process command and arguments.

        Args:
            _command (str): short command string
            args (list|None, optional): command arguments. Defaults to None.

        Raises:
            SystemExit: not git.
            SystemExit: short command not right.
        """

        option: Optional[dict] = self.cmds.get(_command, None)

        # Invalid, if need suggest.
        if option is None:
            print("Don't support this command, please try ", end="")
            color_print("g --show-commands", TermColor.Gold)

            if self.use_recommend:  # check config.
                predicted_command = similar_command(_command, self.cmds.keys())
                if confirm(
                    "%s The wanted command is %s ?[y/n]:"
                    % (
                        Emotion.Icon_Thinking,
                        TermColor.Green + predicted_command + Fx.reset,
                    )
                ):
                    self.process_command(predicted_command, args=args)

            return None

        command = option.get("command", None)
        # Has no command can be executed.
        if not command:
            color_print(
                "Invalid custom short command, nothing can to exec.", TermColor.Red
            )
            return None

        if not option.get("has_arguments", False):
            if args:
                color_print(
                    "The command does not accept parameters. Discard {0}.".format(args),
                    TermColor.Red,
                )
                args = []

        _type = option.get("type", "string")
        if _type == "func":
            try:
                command(args)
            except TermError as e:
                color_print(str(e), TermColor.Red)
        else:  # is string.
            if args:
                args_str = " ".join(args)
                command = " ".join([command, args_str])
            if self.show_original:
                print(
                    "{0}  {1}".format(
                        Emotion.Icon_Rainbow, self.color_command(command)
                    ),
                )
            run_cmd(command)

    ################################
    # Print command help message.
    ################################
    def _generate_help_by_key(self, _key: str, use_color: bool = True) -> str:
        """Generate one help by given key.

        Args:
            _key (str): Short command string.
            use_color (bool, optional): Whether color help message. Defaults to True.

        Returns:
            (str): Help message of one command.
        """

        _msg: str = "    {key_color}{:<9}{reset}{}{command_color}{}{reset}"
        if use_color:
            _key_color = TermColor.Green
            _command_color = TermColor.Gold
        else:
            _key_color = _command_color = ""

        # Get help message and command.
        _help: str = self.cmds[_key].get("help", "")
        try:
            # must have key `command`.
            _command = self.cmds[_key]["command"]
        except:
            raise ValueError("The `command` key can not be empty.")

        # Process help.
        _help = _help + "\n" if _help else ""

        # Process command.
        if callable(_command):
            _command = "Callable: %s" % _command.__name__

        _command = shorten(_command, 70, placeholder="...")
        _command = " " * 13 + _command if _help else _command

        # Splicing and return.
        return _msg.format(
            _key,
            _help,
            _command,
            key_color=_key_color,
            command_color=_command_color,
            reset=Fx.reset,
        )

    def command_help(self) -> None:
        """Print help message."""
        print("These are short commands that can replace git operations:")
        for key in self.cmds.keys():
            msg = self._generate_help_by_key(key)
            print(msg)

    def command_help_by_type(self, command_type: str) -> None:
        """Print a part of help message.

        Print the help information of the corresponding part according to the
        incoming command type string. If there is no print error prompt for the
        type.

        Args:
            command_type (str): A command type of `TYPE`.
        """

        # Process received type.
        command_type = command_type.capitalize().strip()

        # Checking the type whether right.
        if command_type not in CommandType.__members__:
            color_print("There is no such type.", TermColor.Red)
            print("Please use `", end="")
            color_print("g --types", TermColor.Green, end="")
            print(
                "` to view the supported types.",
            )
            if self.use_recommend:
                predicted_type = similar_command(
                    command_type, CommandType.__members__.keys()
                )
                if confirm(
                    "%s The wanted type is %s ?[y/n]:"
                    % (
                        Emotion.Icon_Thinking,
                        TermColor.Green + predicted_type + Fx.reset,
                    )
                ):
                    self.command_help_by_type(predicted_type)
            return None

        # Print help.
        print("These are the orders of {0}".format(command_type))
        for k, v in self.cmds.items():
            belong = v.get("belong", CommandType.Extra)
            # Prevent the `belong` attribute from being set in the custom command.
            if isinstance(belong, CommandType) and belong.value == command_type:
                msg = self._generate_help_by_key(k)
                print(msg)

    @classmethod
    def type_help(cls) -> None:
        """Print all command types with random color."""
        for member in CommandType:
            print(
                "{0}{1}  ".format(
                    Color.fg(
                        random.randint(70, 255),
                        random.randint(70, 255),
                        random.randint(70, 255),
                    ),
                    member.value,
                ),
                end="",
            )
        print(Fx.reset)



# -*- coding:utf-8 -*-
import re


class Fx(object):
    """Text effects
    * trans(string: str): Replace whitespace with escape move right to not
        overwrite background behind whitespace.
    * uncolor(string: str) : Removes all 24-bit color and returns string .
    """

    hide_cursor = "\033[?25l"  # * Hide terminal cursor
    show_cursor = "\033[?25h"  # * Show terminal cursor
    alt_screen = "\033[?1049h"  # * Switch to alternate screen
    normal_screen = "\033[?1049l"  # * Switch to normal screen
    clear_ = "\033[2J\033[0;0f"  # * Clear screen and set cursor to position 0,0
    # * Enable reporting of mouse position on click and release
    mouse_on = "\033[?1002h\033[?1015h\033[?1006h"
    mouse_off = "\033[?1002l"  # * Disable mouse reporting
    # * Enable reporting of mouse position at any movement
    mouse_direct_on = "\033[?1003h"
    mouse_direct_off = "\033[?1003l"  # * Disable direct mouse reporting

    start = "\033["  # * Escape sequence start
    sep = ";"  # * Escape sequence separator
    end = "m"  # * Escape sequence end
    # * Reset foreground/background color and text effects
    reset = rs = "\033[0m"
    bold = b = "\033[1m"  # * Bold on
    unbold = ub = "\033[22m"  # * Bold off
    dark = d = "\033[2m"  # * Dark on
    undark = ud = "\033[22m"  # * Dark off
    italic = i = "\033[3m"  # * Italic on
    unitalic = ui = "\033[23m"  # * Italic off
    underline = u = "\033[4m"  # * Underline on
    ununderline = uu = "\033[24m"  # * Underline off
    blink = bl = "\033[5m"  # * Blink on
    unblink = ubl = "\033[25m"  # * Blink off
    strike = s = "\033[9m"  # * Strike / crossed-out on
    unstrike = us = "\033[29m"  # * Strike / crossed-out off

    # * Precompiled regex for finding a 24-bit color escape sequence in a string
    color_re = re.compile(r"\033\[\d+;\d?;?\d*;?\d*;?\d*m")
    style_re = re.compile(r"\033\[\d+m")

    @staticmethod
    def trans(string):
        return string.replace(" ", "\033[1C")

    @classmethod
    def uncolor(cls, string):
        return cls.color_re.sub("", string)

    @classmethod
    def pure(cls, string):
        return cls.style_re.sub("", cls.uncolor(string))


class Cursor:
    """Class with collection of cursor movement functions:
    Functions:
        .t[o](line, column)
        .r[ight](columns)
        .l[eft](columns)
        .u[p](lines)
        .d[own](lines)
        .save()
        .restore()
    """

    @staticmethod
    def to(line, col):
        # * Move cursor to line, column
        return "\033[{};{}f".format(line, col)

    @staticmethod
    def right(dx):
        return "\033[{}C".format(dx)

    @staticmethod
    def left(dx):
        return "\033[{}D".format(dx)

    @staticmethod
    def up(dy):
        return "\033[{}A".format(dy)

    @staticmethod
    def down(dy):
        return "\033[{}B".format(dy)

    save = "\033[s"  # * Save cursor position
    restore = "\033[u"  # * Restore saved cursor postion
    t = to
    r = right
    l = left
    u = up
    d = down

    @staticmethod
    def hide_cursor():
        print("\033[?25l")  # * Hide terminal cursor

    @staticmethod
    def show_cursor():
        print("\033[?25h")  # * Show terminal cursor



# -*- coding:utf-8 -*-

from copy import deepcopy

from .style import Symbol
from .escape import Fx


class baseTable(object):

    """Docstring for baseTable."""

    table_str_list: list
    each_max: list

    def __init__(self, frame_format: str, nil: str, title: str = ""):
        if frame_format not in Symbol.rune.keys():
            frame_format = "bold"
        self.rune = Symbol.rune[frame_format]
        self.nil = nil
        self.title = title

        # create table when init.
        self.fix_data()
        self.create_table()

    def __str__(self):
        if getattr(self, "table_str_list"):
            return "\n".join(self.table_str_list)
        else:
            return ""

    def _try_append_title(self):
        if self.title:
            line_max = sum(self.each_max) + len(self.each_max) + 1
            _t = "{}{:^%s}{}" % line_max
            self.table_str_list.append(_t.format(Fx.b, self.title, Fx.ub))

    def fix_data(self):
        raise NotImplementedError()

    def create_table(self):
        raise NotImplementedError()

    def print(self):
        for each in self.table_str_list:
            print(each)


class Table(baseTable):
    """Create a table from list.

    data format:
        title = ["name", "age", "gender"]
        data = [
            ["bob", "20", "f"],
            ["tom", "19", "f"],
        ]

    table format:
        ┏━━━━┳━━━┳━━━━━━┓
        ┃name┃age┃gender┃
        ┣━━━━╋━━━╋━━━━━━┫
        ┃bob ┃20 ┃f     ┃
        ┃tom ┃19 ┃f     ┃
        ┗━━━━┻━━━┻━━━━━━┛

    """

    def __init__(
        self,
        header: list,
        data: list,
        title: str = "",
        frame_format: str = "bold",
        nil: str = "",
    ):
        # Check data.
        if not isinstance(header, list):
            raise TypeError("title need is a list.")
        self.header = deepcopy(header)

        # Check data.
        if not isinstance(data, list):
            raise TypeError("data need is a list.")
        for item in data:
            if not isinstance(item, list):
                raise TypeError("each item of data need is a list.")
        self.data = deepcopy(data)

        super().__init__(frame_format, nil, title=title)

    def fix_data(self):
        self.header_len = header_len = len(self.header)
        self.each_max = each_max = [len(i) for i in self.header]

        for item in self.data:
            # Complete missing element.
            item_len = len(item)
            if item_len < header_len:
                item.extend([self.nil] * (header_len - item_len))
            # Calc each max.
            for i, x in enumerate(item):
                x_len = len(Fx.pure(x))
                each_max[i] = max(each_max[i], x_len)

    def create_table(self):
        each_max = self.each_max
        rune = self.rune

        _t = "%s{:<%s}"
        _template = (
            "".join([_t % (self.rune[1], number) for number in each_max]) + rune[1]
        )

        _head = [
            rune[2] + rune[-3].join([rune[0] * i for i in each_max]) + rune[3],
            _template.format(*self.header),
            rune[6] + rune[-1].join([rune[0] * i for i in each_max]) + rune[7],
        ]

        _body = [_template.format(*item) for item in self.data]

        _foot = rune[4] + rune[-2].join([rune[0] * i for i in each_max]) + rune[5]

        self.table_str_list = []
        self._try_append_title()
        self.table_str_list.extend(_head)
        self.table_str_list.extend(_body)
        self.table_str_list.append(_foot)


class dTable(baseTable):
    """Create table from a special format dict.

    d_data format: dict[str, dict[str, str]]
        d_data = {
            'Fruit color': {
                'apple': 'red',
                'grape': 'purple',
            },
            'Animal color': {
                'cattle': 'yellow',
                'sheep': 'white',
            },
        }

    table format:
        ┏━━━━━━━━━━━━━┓
        ┃ Fruit color ┃
        ┣━━━━━━┳━━━━━━┫
        ┃apple ┃red   ┃
        ┃grape ┃purple┃
        ┣━━━━━━┻━━━━━━┫
        ┃Animal color ┃
        ┣━━━━━━┳━━━━━━┫
        ┃cattle┃yellow┃
        ┃sheep ┃white ┃
        ┣━━━━━━┻━━━━━━┫
        ┃      ────END┃
        ┗━━━━━━━━━━━━━┛

    """

    def __init__(
        self, d_data: dict, title: str = "", frame_format: str = "bold", nil: str = ""
    ):
        # check data whether right.
        if not isinstance(d_data, dict):
            raise TypeError("d_data need is a dict.")
        for item in d_data.values():
            if not isinstance(item, dict):
                raise TypeError("the item of d_data need is a dict.")

        self.data = deepcopy(d_data)

        super().__init__(frame_format, nil, title=title)

    def fix_data(self):
        self.each_max = each_max = [0, 0]

        max_core_len = 0

        for core, sub_dict in self.data.items():
            max_core_len = max(max_core_len, len(core))
            for k, v in sub_dict.items():
                each_max[0] = max(each_max[0], len(Fx.pure(k)))
                each_max[1] = max(each_max[1], len(Fx.pure(v)))

        # for Ensure that the table is output correctly when the len of sub title
        # bigger than the len of item.
        sum_each_max = sum(each_max)
        if max_core_len > sum_each_max:
            each_max[1] += max_core_len - sum_each_max

        self.line_max = sum(each_max) + len(each_max) + 1

    def create_table(self):
        rune = self.rune
        each_max = self.each_max
        line_max = self.line_max

        tb = self.table_str_list = []
        self._try_append_title()

        _core_template = "%(flag)s{:^%(number)s}%(flag)s" % {
            "flag": rune[1],
            "number": line_max - 2,
        }
        _item_template = "%(flag)s{:<%(num1)s}%(flag)s{:<%(num2)s}%(flag)s" % {
            "flag": rune[1],
            "num1": each_max[0],
            "num2": each_max[1],
        }
        _end_template = "%(flag)s{:>%(number)s}%(flag)s" % {
            "flag": rune[1],
            "number": line_max - 2,
        }

        tb.append(rune[2] + rune[0] * (line_max - 2) + rune[3])

        for key, sub_dict in self.data.items():
            tb.append(_core_template.format(key))
            tb.append(
                rune[6] + rune[-3].join([rune[0] * i for i in each_max]) + rune[7]
            )
            for k, v in sub_dict.items():
                tb.append(_item_template.format(k, v))
            tb.append(
                rune[6] + rune[-2].join([rune[0] * i for i in each_max]) + rune[7]
            )

        tb.append(_end_template.format("────END"))
        tb.append(rune[4] + rune[0] * (line_max - 2) + rune[5])



# -*- coding:utf-8 -*-

import os
import re
import shutil
import textwrap
import logging
from typing import Optional

from .utils import exec_cmd, color_print
from .common import Fx, TermColor, Symbol
from .common.str_table import dTable

Log = logging.getLogger(__name__)


def git_version() -> str:
    """Get Git version."""
    _, git_version_ = exec_cmd("git --version")
    if git_version_:
        return git_version_
    else:
        return ""


# Not detected, the result is None
Git_Version: str = git_version()


def current_repository() -> str:
    """Get the current git repository path. If not, the path is empty."""
    err, path = exec_cmd("git rev-parse --git-dir")

    if err:
        return ""

    path = path.strip()
    if path == ".git":
        repository_path = os.getcwd()
    else:
        repository_path = path[:-5]
    return repository_path


REPOSITORY_PATH: str = current_repository()
IS_GIT_REPOSITORY: bool = True if REPOSITORY_PATH else False


def _config_normal_output(conf: dict[str, dict]):
    for t, d in conf.items():
        print(TermColor.Red + f"[{t}]")
        for k, v in d.items():
            print(" " * 4 + TermColor.SkyBlue + k, end="")
            print("=" + Fx.italic + TermColor.MediumVioletRed + v + Fx.reset)


def _config_table_output(conf: dict[str, dict]):
    for sub in conf.values():
        for k, v in sub.items():
            sub[k] = f"{TermColor.Green}{v:<40}{Fx.rs}"

    tb = dTable(conf, title='Git Local Config')
    tb.print()


def output_git_local_config(style: str = "table") -> None:
    """Print the local config of current git repository."""

    if not IS_GIT_REPOSITORY:
        color_print("This directory is not a git repository yet.", TermColor.Red)
        return None

    _re = re.compile(r"\w+\s=\s.*?")
    width, _ = shutil.get_terminal_size()
    if width < 72:
        style = "normal"

    try:
        with open(REPOSITORY_PATH + "/.git/config", "r") as cf:
            context = cf.read()
    except Exception as e:
        color_print(
            "Error reading configuration file. {0}".format(str(e)), TermColor.Red
        )
    else:
        l = re.split(r"\r\n|\r|\n", context)
        l_len = len(l)
        config_dict = {}
        idx = 0

        while idx < l_len:
            if l[idx].startswith("["):
                config_type = l[idx][1:-1].strip()
                if not config_dict.get(config_type):
                    config_dict[config_type] = {}
                    idx += 1
                    while idx < l_len and not l[idx].startswith("["):
                        if "=" not in l[idx]:
                            idx += 1
                            continue
                        key, value = l[idx].split("=", 1)
                        config_dict[config_type][key.strip()] = value.strip()
                        idx += 1
        # print(config_dict)

        if style == "normal":
            _config_normal_output(config_dict)
        elif style == "table":
            _config_table_output(config_dict)
        else:
            _config_normal_output(config_dict)


def output_repository_info(
    show_path: bool = True,
    show_remote: bool = True,
    show_branches: bool = True,
    show_lastest_log: bool = True,
    show_summary: bool = True,
) -> None:
    """Print some information of the repository.

    repository: `Repository_Path`
    remote: read from '.git/conf'
    >>> all_branch = run_cmd_with_resp('git branch --all --color')
    >>> lastest_log = run_cmd_with_resp('git log -1')
    """

    print("waiting ...", end="")

    error_str = TermColor.Red + "Error getting" + Fx.reset

    # Print content.
    print("\r[%s]        \n" % (Fx.b + "Repository Information" + Fx.reset,))
    if show_path:
        print(
            "Repository: \n\t%s\n" % (TermColor.SkyBlue + REPOSITORY_PATH + Fx.reset,)
        )

    # Get remote url.
    if show_remote:
        try:
            with open(REPOSITORY_PATH + "/.git/config", "r") as cf:
                config = cf.read()
        except Exception:
            remote = error_str
        else:
            res = re.findall(r"url\s=\s(.*)", config)
            remote = "\n".join(
                [
                    "\t%s%s%s%s" % (Fx.italic, TermColor.SkyBlue, x, Fx.reset)
                    for x in res
                ]
            )
        print("Remote: \n%s\n" % remote)

    # Get all branches.
    if show_branches:
        err, res = exec_cmd("git branch --all --color")
        if err:
            branches = "\t" + error_str
        else:
            branches = textwrap.indent(res, "\t")
        print("Branches: \n%s\n" % branches)

    # Get the lastest log.
    if show_lastest_log:
        err, res = exec_cmd("git log --stat --oneline --decorate -1 --color")
        if err:
            git_log = "\t" + error_str
        else:
            # git_log = "\n".join(["\t" + x for x in res.strip().split("\n")])
            git_log = textwrap.indent(res, "\t")
        print("Lastest log:\n%s\n" % git_log)

    # Get git summary.
    if show_summary:
        err, res = exec_cmd("git shortlog --summary --numbered")
        if err:
            summary = "\t" + error_str
        else:
            summary = textwrap.indent(res, "\t")
        print("Summary:\n%s\n" % summary)



# -*- coding:utf-8 -*-

import os
import logging
import logging.handlers


class LogHandle(object):
    """Set log handle.
    Attributes:
        FMT_NORMAL: Log style in normal mode.
        FMT_DEBUG: Log style in debug mode.

    Functions:
        setup_logging: setup log handle setting.

    Raises:
        SystemExit: When the log file cannot be written.
    """

    FMT_NORMAL = logging.Formatter(
        fmt="%(asctime)s %(levelname).4s %(message)s", datefmt="%H:%M:%S"
    )
    FMT_DEBUG = logging.Formatter(
        fmt="%(asctime)s.%(msecs)03d %(levelname).4s [%(name)s] %(message)s",
        datefmt="%H:%M:%S",
    )

    @classmethod
    def setup_logging(cls, debug=False, log_file=None):
        root_logger = logging.getLogger()

        if debug:
            log_level = logging.DEBUG
            formatter = cls.FMT_DEBUG
        else:
            log_level = logging.INFO
            formatter = cls.FMT_NORMAL

        if log_file:
            if log_file is None:
                log_handle = logging.StreamHandler()
            else:
                dir_path = os.path.dirname(log_file)
                if not os.path.isdir(dir_path):
                    os.makedirs(dir_path, exist_ok=True)
                try:
                    log_handle = logging.handlers.RotatingFileHandler(
                        log_file, maxBytes=1048576, backupCount=4
                    )
                except PermissionError:
                    print("No permission to write to '{0}' directory!".format(log_file))
                    raise SystemExit(1)

        log_handle.setFormatter(formatter)
        log_handle.setLevel(log_level)

        root_logger.addHandler(log_handle)
        root_logger.setLevel(0)



# -*- coding:utf-8 -*-

# MIT License
#
# Copyright (c) 2021 Zachary
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


__project__ = "pigit"
__version__ = "1.3.1"
__url__ = "https://github.com/zlj-zz/pigit.git"
__uri__ = __url__

__author__ = "Zachary Zhang"
__email__ = "zlj19971222@outlook.com"

__license__ = "MIT"
__copyright__ = "Copyright (c) 2021 Zachary"


import os
import sys
import argparse
import logging
import textwrap
from distutils.util import strtobool
from shutil import get_terminal_size
from typing import Optional

from .log import LogHandle
from .utils import confirm, color_print, is_color
from .common import Color, Fx, TermColor
from .git_utils import (
    Git_Version,
    REPOSITORY_PATH,
    output_repository_info,
    output_git_local_config,
)
from .decorator import time_it
from .codecounter import CodeCounter
from .shell_completion import ShellCompletion, process_argparse
from .gitignore import GitignoreGenetor
from .command_processor import GitProcessor, Git_Cmds, CommandType


Log = logging.getLogger(__name__)

#####################################################################
# Part of compatibility.                                            #
# Handled the incompatibility between python2 and python3.          #
#####################################################################

# For windows.
IS_WIN: bool = sys.platform.lower().startswith("win")
Log.debug("Runtime platform is windows: {0}".format(IS_WIN))
if IS_WIN:
    USER_HOME = os.environ["USERPROFILE"]
    PIGIT_HOME = os.path.join(USER_HOME, __project__)
else:
    USER_HOME = os.environ["HOME"]
    PIGIT_HOME = os.path.join(USER_HOME, ".config", __project__)

LOG_PATH = PIGIT_HOME + "/log/{0}.log".format(__project__)
COUNTER_PATH = PIGIT_HOME + "/Counter"


#####################################################################
# Configuration.                                                    #
#####################################################################
class ConfigError(Exception):
    """Config error. Using by `Config`."""

    pass


class Config(object):
    """PIGIT configuration class."""

    _conf_path: str = PIGIT_HOME + "/pigit.conf"  # default config path.

    CONFIG_TEMPLATE: str = textwrap.dedent(
        """\
        #? Config file for pigit v. {version}

        #  ____ ___ ____ ___ _____                            __ _
        # |  _ \\_ _/ ___|_ _|_   _|           ___ ___  _ __  / _(_) __ _
        # | |_) | | |  _ | |  | |_____ _____ / __/ _ \\| '_ \\| |_| |/ _` |
        # |  __/| | |_| || |  | |_____|_____| (_| (_) | | | |  _| | (_| |
        # |_|  |___\\____|___| |_|            \\___\\___/|_| |_|_| |_|\\__, |
        #                                     {version:>20} |___/
        # Git-tools -- pigit configuration.

        # Show original git command.
        gitprocessor_show_original={gitprocessor_show_original}

        # Is it recommended to correct when entering wrong commands.
        gitprocessor_use_recommend={gitprocessor_use_recommend}

        # Whether color is enabled in interactive mode.
        gitprocessor_interactive_color={gitprocessor_interactive_color}

        # Display time of help information in interactive mode, 0 is permanent.
        gitprocessor_interactive_help_showtime={gitprocessor_interactive_help_showtime}

        # Whether to use the ignore configuration of the `.gitignore` file.
        codecounter_use_gitignore={codecounter_use_gitignore}

        # Whether show files that cannot be counted.
        codecounter_show_invalid={codecounter_show_invalid}

        # Whether show files icons. Font support required, like: 'Nerd Font'
        codecounter_show_icon={codecounter_show_icon}

        # Output format of statistical results.
        # Supported: [table, simple]
        # When the command line width is not enough, the `simple ` format is forced.
        codecounter_result_format={codecounter_result_format}

        # Timeout for getting `.gitignore` template from net.
        gitignore_generator_timeout={gitignore_generator_timeout}

        # Git local config print format.
        # Supported: [table, normal]
        git_config_format={git_config_format}

        # Control which parts need to be displayed when viewing git repository information.
        repository_show_path={repository_show_path}
        repository_show_remote={repository_show_remote}
        repository_show_branchs={repository_show_branchs}
        repository_show_lastest_log={repository_show_lastest_log}
        repository_show_summary={repository_show_summary}

        # Whether with color when use `-h` get help message.
        help_use_color={help_use_color}

        # The max line width when use `-h` get help message.
        help_max_line_width={help_max_line_width}

        # Whether run PIGIT in debug mode.
        debug_mode={debug_mode}

        """
    )

    # yapf: disable
    _keys :list[str]= [
        'gitprocessor_show_original', 'gitprocessor_use_recommend',
        'gitprocessor_interactive_color', 'gitprocessor_interactive_help_showtime',
        'codecounter_use_gitignore', 'codecounter_show_invalid',
        'codecounter_result_format', 'codecounter_show_icon',
        'gitignore_generator_timeout',
        'git_config_format',
        'repository_show_path', 'repository_show_remote', 'repository_show_branchs',
        'repository_show_lastest_log', 'repository_show_summary',
        'help_use_color', 'help_max_line_width',
        'debug_mode'
    ]
    # yapf: enable

    # config default values.
    gitprocessor_show_original: bool = True
    gitprocessor_use_recommend: bool = True
    gitprocessor_interactive_color: bool = True
    gitprocessor_interactive_help_showtime: float = 1.5

    codecounter_use_gitignore: bool = True
    codecounter_show_invalid: bool = False
    codecounter_show_icon: bool = False
    codecounter_result_format: str = "table"  # table, simple
    _supported_result_format: list = ["table", "simple"]

    gitignore_generator_timeout: int = 60

    git_config_format: str = "table"
    _supported_git_config_format: list = ["normal", "table"]

    repository_show_path: bool = True
    repository_show_remote: bool = True
    repository_show_branchs: bool = True
    repository_show_lastest_log: bool = True
    repository_show_summary: bool = False

    help_use_color: bool = True
    help_max_line_width: int = 90

    debug_mode: bool = False

    # Store warning messages.
    warnings: list = []

    def __init__(self, path: Optional[str] = None) -> None:
        super(Config, self).__init__()
        if not path:
            self.config_path = self._conf_path
        else:
            self.config_path = path
        conf = self.load_config()

        for key in self._keys:
            if key in conf.keys() and conf[key] != "==error==":
                setattr(self, key, conf[key])

    def load_config(self) -> dict:
        new_config = {}
        config_file = self.config_path
        try:
            with open(config_file) as cf:
                for line in cf:
                    line = line.strip()
                    if line.startswith("#? Config"):
                        new_config["version"] = line[line.find("v. ") + 3 :]
                        continue
                    if line.startswith("#"):
                        # comment line.
                        continue
                    if "=" not in line:
                        # invalid line.
                        continue
                    key, line = line.split("=", maxsplit=1)

                    # processing.
                    key = key.strip()
                    line = line.strip().strip('"')

                    # checking.
                    if key not in self._keys:
                        self.warnings.append("'{0}' is not be supported!".format(key))
                        continue
                    elif type(getattr(self, key)) == int:
                        try:
                            new_config[key] = int(line)
                        except ValueError:
                            self.warnings.append(
                                'Config key "{0}" should be an integer!'.format(key)
                            )
                    elif type(getattr(self, key)) == bool:
                        try:
                            # True values are y, yes, t, true, on and 1;
                            # false values are n, no, f, false, off and 0.
                            # Raises ValueError if val is anything else.
                            new_config[key] = bool(strtobool(line))
                        except ValueError:
                            self.warnings.append(
                                'Config key "{0}" can only be True or False!'.format(
                                    key
                                )
                            )
                    elif type(getattr(self, key)) == str:
                        if "color" in key and not is_color(line):
                            self.warnings.append(
                                'Config key "{0}" should be RGB string, like: #FF0000'.format(
                                    key
                                )
                            )
                            continue
                        new_config[key] = str(line)
        except Exception as e:
            Log.error(str(e) + str(e.__traceback__))

        if (  # check codecounter output format whether supported.
            "codecounter_result_format" in new_config
            and new_config["codecounter_result_format"]
            not in self._supported_result_format
        ):
            new_config["codecounter_result_format"] = "==error=="
            self.warnings.append(
                'Config key "{0}" support must in {1}'.format(
                    "codecounter_result_format", self._supported_result_format
                )
            )

        if (
            "git_config_format" in new_config
            and new_config["git_config_format"] not in self._supported_git_config_format
        ):
            new_config["git_config_format"] = "==error=="
            self.warnings.append(
                'Config key "{0}" support must in {1}'.format(
                    "git_config_format", self._supported_git_config_format
                )
            )

        if "version" in new_config and not (
            # If current version is a [beta] verstion then will not tip.
            # Else if version is not right will tip.
            new_config["version"] == __version__
            or "beta" in __version__
            or "alpha" in __version__
        ):
            print(new_config["version"])
            self.warnings.append(
                "The current configuration file is not up-to-date."
                "You'd better recreate it."
            )

        return new_config

    def create_config_template(self) -> None:
        if not os.path.isdir(PIGIT_HOME):
            os.makedirs(PIGIT_HOME, exist_ok=True)

        if os.path.exists(self.config_path) and not confirm(
            "Configuration exists, overwrite? [y/n]"
        ):
            return None

        # Try to load already has config.
        conf = self.load_config()
        for key in self._keys:
            if not conf.get(key, None) or conf[key] == "==error==":
                conf[key] = getattr(self, key)
        conf["version"] = __version__

        # Write config. Will save before custom setting.
        try:
            with open(
                self.config_path, "w" if os.path.isfile(self.config_path) else "x"
            ) as f:
                f.write(self.CONFIG_TEMPLATE.format(**conf))
        except Exception as e:
            Log.error(str(e) + str(e.__traceback__))
            print("Failed, create config.")
        else:
            print("Successful.")


CONFIG = Config()
if CONFIG.warnings:
    for warning in CONFIG.warnings:
        print(warning)
    CONFIG.warnings = []


#####################################################################
# Implementation of additional functions.                           #
#####################################################################
def introduce() -> None:
    """Print the description information."""

    # Print tools version and path.
    color_print("[%s] version: %s" % (__project__, __version__), Fx.b)

    # Print git version.
    if Git_Version is None:
        print("Don't found Git, maybe need install.")
    else:
        print(Git_Version)

    # Print package path.
    color_print("Path: ", Fx.b, end="")
    color_print("%s\n" % __file__, TermColor.SkyBlue, Fx.underline)

    # Print description.
    color_print("Description:", Fx.b)
    color_print(
        (
            "  Terminal tool, help you use git more simple."
            " Support Linux and MacOS. Partial support for windows.\n"
            "  It use short command to replace the original command, like: \n"
            "  `pigit ws` -> `git status --short`, `pigit b` -> `git branch`.\n"
            "  Also you use `pigit -s` to get the all short command, have fun"
            " and good lucky.\n"
            "  The open source path: %s" % (TermColor.SkyBlue + Fx.underline + __url__)
        ),
        Fx.italic,
    )

    print("\nYou can use ", end="")
    color_print("-h", TermColor.Green, end="")
    print(" or ", end="")
    color_print("--help", TermColor.Green, end="")
    print(" to get help and more usage.\n")


def get_extra_cmds() -> dict:
    """Get custom cmds.

    Load the `extra_cmds.py` file under PIGIT HOME, check whether `extra_cmds`
    exists, and return it. If not have, return a empty dict.

    Returns:
        (dict[str,str]): extra cmds dict.
    """
    import imp

    extra_cmd_path: str = PIGIT_HOME + "/extra_cmds.py"
    extra_cmds = {}

    if os.path.isfile(extra_cmd_path):
        try:
            extra_cmd = imp.load_source("extra_cmd", extra_cmd_path)
        except Exception as e:
            Log.error(
                "Can't load file '{0}';{1};{2}".format(
                    extra_cmd_path, str(e), str(e.__traceback__)
                )
            )
        else:
            try:
                extra_cmds = extra_cmd.extra_cmds
            except AttributeError:
                Log.error("Can't found dict name is 'extra_cmds'.")

    # print(extra_cmds)
    return extra_cmds


class CustomHelpFormatter(argparse.HelpFormatter):
    """Formatter for generating usage messages and argument help strings.

    This class inherits `argparse.HelpFormatter` and rewrites some methods
    to complete customization.
    """

    def __init__(
        self, prog, indent_increment=2, max_help_position=24, width=90, colors=[]
    ):
        width = CONFIG.help_max_line_width
        max_width, _ = get_terminal_size()

        width = width if width < max_width else max_width - 2
        super(CustomHelpFormatter, self).__init__(
            prog, indent_increment, max_help_position, width
        )
        if not colors or not isinstance(colors, list):
            colors = {
                "red": Color.fg("#FF6347"),  # Tomato
                "green": Color.fg("#98FB98"),  # PaleGreen
                "yellow": Color.fg("#EBCB8C"),  # Yellow
                "blue": Color.fg("#87CEFA"),  # SkyBlue
                # Color.fg("#FFC0CB"),  # Pink
            }
        self.colors = colors
        self.color_len = len(colors)
        self._old_color = None

    def _format_action(self, action):
        # determine the required width and the entry label
        help_position = min(self._action_max_length + 2, self._max_help_position)
        help_width = max(self._width - help_position, 11)
        action_width = help_position - self._current_indent - 2
        action_header = self._format_action_invocation(action)

        # no help; start on same line and add a final newline
        if not action.help:
            tup = self._current_indent, "", action_header
            action_header = "%*s%s\n" % tup

        # short action name; start on the same line and pad two spaces
        elif len(action_header) <= action_width:
            tup = self._current_indent, "", action_width, action_header
            action_header = "%*s%-*s  " % tup
            indent_first = 0

        # long action name; start on the next line
        else:
            tup = self._current_indent, "", action_header
            action_header = "%*s%s\n" % tup
            indent_first = help_position

        # collect the pieces of the action help
        # @Overwrite
        if CONFIG.help_use_color:
            _color = self.colors["green"]
        else:
            _color = ""
        parts = [_color + action_header + Fx.reset]

        # if there was help for the action, add lines of help text
        if action.help:
            help_text = self._expand_help(action)
            help_lines = self._split_lines(help_text, help_width)
            parts.append("%*s%s\n" % (indent_first, "", help_lines[0]))
            for line in help_lines[1:]:
                parts.append("%*s%s\n" % (help_position, "", line))

        # or add a newline if the description doesn't end with one
        elif not action_header.endswith("\n"):
            parts.append("\n")

        # if there are any sub-actions, add their help as well
        for subaction in self._iter_indented_subactions(action):
            parts.append(self._format_action(subaction))

        # return a single string
        return self._join_parts(parts)

    def _fill_text(self, text, width, indent):
        try:
            return "".join(indent + line for line in text.splitlines(keepends=True))
        except TypeError:
            return "".join(indent + line for line in text.splitlines())

    def _get_help_string(self, action):
        help = action.help
        if "%(default)" not in action.help:
            if action.default is not argparse.SUPPRESS:
                defaulting_nargs = [argparse.OPTIONAL, argparse.ZERO_OR_MORE]
                if action.option_strings or action.nargs in defaulting_nargs:
                    # help += " (default: %(default)s)"
                    pass
        return help


class Parser(object):
    def __init__(self) -> None:
        super(Parser, self).__init__()
        self._parser = argparse.ArgumentParser(
            prog="pigit",
            description="If you want to use some original git commands, please use -- to indicate.",
            prefix_chars="-",
            formatter_class=CustomHelpFormatter,
        )
        self._add_arguments()

    def _add_arguments(self) -> None:
        self._parser.add_argument(
            "-s",
            "--show-commands",
            action="store_true",
            help="List all available short command and wealth and exit.",
        )
        self._parser.add_argument(
            "-S",
            "--show-command",
            type=str,
            metavar="TYPE",
            dest="command_type",
            help="According to given type(%s) list available short command and wealth and exit."
            % ", ".join(CommandType.__members__.keys()),
        )
        self._parser.add_argument(
            "-t",
            "--types",
            action="store_true",
            help="List all command types and exit.",
        )
        self._parser.add_argument(
            "-f",
            "--config",
            action="store_true",
            help="Display the config of current git repository and exit.",
        )
        self._parser.add_argument(
            "-i",
            "--information",
            action="store_true",
            help="Show some information about the current git repository.",
        )
        self._parser.add_argument(
            "-c",
            "--count",
            nargs="?",
            const=".",
            type=str,
            metavar="PATH",
            help="Count the number of codes and output them in tabular form."
            "A given path can be accepted, and the default is the current directory.",
        )
        self._parser.add_argument(
            "-C",
            "--complete",
            action="store_true",
            help="Add shell prompt script and exit.(Supported `bash`, `zsh`, `fish`)",
        )
        self._parser.add_argument(
            "--create-ignore",
            type=str,
            metavar="TYPE",
            dest="ignore_type",
            help="Create a demo `.gitignore` file. Need one argument, support: [%s]"
            % ", ".join(GitignoreGenetor.Supported_Types.keys()),
        )
        self._parser.add_argument(
            "--create-config",
            action="store_true",
            help="Create a preconfigured file of PIGIT."
            "(If a profile exists, the values available in it are used)",
        )
        self._parser.add_argument(
            "--shell",
            action="store_true",
            help="Go to the pigit shell mode.",
        )
        self._parser.add_argument(
            "--debug",
            action="store_true",
            help="Current runtime in debug mode.",
        )
        self._parser.add_argument(
            "--out-log",
            action="store_true",
            help="Print log to console.",
        )
        self._parser.add_argument(
            "-v",
            "--version",
            action="version",
            help="Show version and exit.",
            version="Version: %s" % __version__,
        )
        self._parser.add_argument(
            "command", nargs="?", type=str, help="Short git command."
        )
        self._parser.add_argument(
            "args", nargs="*", type=str, help="Command parameter list."
        )

    def parse(self, custom_commands: Optional[list] = None):
        if custom_commands:
            return self._parser.parse_known_args(custom_commands)
        return self._parser.parse_known_args()

    def process(self, known_args, extra_unknown: Optional[list] = None) -> None:
        try:
            if known_args.complete:
                # Generate competion vars dict.
                completion_vars = {
                    key: value.get("help", "") for key, value in Git_Cmds.items()
                }

                # Update var dict with shell command.
                completion_vars.update(process_argparse(self._parser))

                ShellCompletion(
                    __project__, completion_vars, PIGIT_HOME
                ).complete_and_use()
                return None

            if known_args.config:
                output_git_local_config(CONFIG.git_config_format)

            if known_args.information:
                output_repository_info(
                    show_path=CONFIG.repository_show_path,
                    show_remote=CONFIG.repository_show_remote,
                    show_branches=CONFIG.repository_show_branchs,
                    show_lastest_log=CONFIG.repository_show_lastest_log,
                    show_summary=CONFIG.repository_show_summary,
                )

            if known_args.create_config:
                return CONFIG.create_config_template()

            if known_args.ignore_type:
                GitignoreGenetor(timeout=CONFIG.gitignore_generator_timeout,).launch(
                    known_args.ignore_type,
                    dir_path=REPOSITORY_PATH,
                )
                return None

            if known_args.count:
                path = known_args.count if known_args.count != "." else os.getcwd()
                CodeCounter(
                    count_path=path,
                    use_ignore=CONFIG.codecounter_use_gitignore,
                    result_saved_path=COUNTER_PATH,
                    result_format=CONFIG.codecounter_result_format,
                    use_icon=CONFIG.codecounter_show_icon,
                ).count_and_format_print(
                    show_invalid=CONFIG.codecounter_show_invalid,
                )
                return None

            git_processor = GitProcessor(
                extra_cmds=get_extra_cmds(),
                use_recommend=CONFIG.gitprocessor_use_recommend,
                show_original=CONFIG.gitprocessor_show_original,
                use_color=CONFIG.gitprocessor_interactive_color,
                help_wait=CONFIG.gitprocessor_interactive_help_showtime,
            )

            def _shell_mode():
                import pigit.tomato

                print(
                    "Welcome come PIGIT shell.\n"
                    "You can use short commands directly. Input '?' to get help.\n"
                )
                while True:
                    command = input("(pigit)> ").strip()
                    if not command:
                        continue

                    command_args = command.split()
                    head_command = command_args[0]

                    if head_command in ["quit", "exit"]:  # ctrl+c
                        break
                    elif head_command in git_processor.cmds.keys():
                        git_processor.process_command(head_command, command_args[1:])
                    elif head_command == "tomato":
                        # Tomato clock.
                        pigit.tomato.main(command_args)
                    elif head_command == "?":
                        other_ = command_args[1:]
                        if not other_:
                            git_processor.command_help()
                        else:
                            print(other_)
                            for item in other_:
                                if item in git_processor.cmds.keys():
                                    print(git_processor._generate_help_by_key(item))
                                elif item == "tomato":
                                    pigit.tomato.help("tomato")
                    else:
                        os.system(command)
                return None

            if known_args.shell:
                _shell_mode()

            if known_args.show_commands:
                return git_processor.command_help()

            if known_args.command_type:
                return git_processor.command_help_by_type(known_args.command_type)

            if known_args.types:
                return git_processor.type_help()

            if known_args.command:
                if known_args.command == "shell":
                    return _shell_mode()
                else:
                    command = known_args.command
                    known_args.args.extend(extra_unknown)  # type: list
                    git_processor.process_command(command, known_args.args)
                    return None

            # Don't have invalid command list.
            if not list(filter(lambda x: x, vars(known_args).values())):
                introduce()
        except (KeyboardInterrupt, EOFError):
            raise SystemExit(0)


@time_it
def main(custom_commands: Optional[list] = None):
    parser = Parser()

    # parse custom comand, if has.
    if custom_commands is not None:
        stdargs = parser.parse(custom_commands)
    stdargs, extra_unknown = parser.parse()
    # print(stdargs, extra_unknown)

    # Setup log handle.
    LogHandle.setup_logging(
        debug=stdargs.debug or CONFIG.debug_mode,
        log_file=None if stdargs.out_log else LOG_PATH,
    )

    # process result.
    parser.process(stdargs, extra_unknown)


if __name__ == "__main__":
    main()



# /usr/bin/env python3
# -*- coding:utf-8 -*-

"""
 For `python -m pigit`.
 When you run a python module directly, the interpreter
 looks for the `__main__.py` file in the module.
"""

from pigit import main

if __name__ == "__main__":
    main()



# -*- coding:utf-8 -*-

import os
import re
import logging
import textwrap
from typing import Optional

from .utils import run_cmd, exec_cmd


Log = logging.getLogger(__name__)


class ShellCompletionError(Exception):
    """ShellCompletion error class."""

    pass


class ShellCompletion(object):
    """Implement and inject help classes for completion scripts.

    Attributes:
        _TEMPLATE_ZSH (str): zsh completion template.
        _TEMPLATE_BASH (str): bash completion template.
        _TEMPLATE_FISH (str): fish completion template.
        Supported_Shell (list): supported shell list.
    """

    # The completion items each line muse be:
    #   -C\:"Add shell prompt script and exit.(Supported `bash`, `zsh`)"\\
    _TEMPLATE_ZSH: str = textwrap.dedent(
        """\
        #compdef %(prop)s

        complete_%(prop)s(){
        local curcontext="$curcontext" state line ret=1
        typeset -A opt_args

        _alternative\\
          \'args:options arg:((\\
            %(complete_vars)s
          ))\'\\
          'files:filename:_files'
        return ret
        }

        compdef complete_%(prop)s %(prop)s
        """
    )

    _TEMPLATE_BASH: str = textwrap.dedent(
        """\
        #!/usr/env bash

        _complete_%(prop)s(){
          if [[ "${COMP_CWORD}" == "1" ]];then
              COMP_WORD="%(complete_vars)s"
              COMPREPLY=($(compgen -W "$COMP_WORD" -- ${COMP_WORDS[${COMP_CWORD}]}))
          fi
        }

        complete -F _complete_%(prop)s %(prop)s
        """
    )

    _TEMPLATE_FISH: str = textwrap.dedent(
        """\
        function complete_%(prop)s;
            set -l response;

            for value in (env %(complete_vars)s=fish_complete COMP_WORDS=(commandline -cp) \
        COMP_CWORD=(commandline -t) %(prop)s);
                set response $response $value;
            end;

            for completion in $response;
                set -l metadata (string split "," $completion);

                if test $metadata[1] = "dir";
                    __fish_complete_directories $metadata[2];
                else if test $metadata[1] = "file";
                    __fish_complete_path $metadata[2];
                else if test $metadata[1] = "plain";
                    print $metadata[2];
                end;
            end;
        end;

        complete --no-files --command %(prop)s --arguments \
        "(complete_%(prop)s)";
        """
    )

    Supported_Shell: list = ["zsh", "bash", "fish"]

    def __init__(
        self,
        prop: str,
        complete_vars: dict,
        script_dir: str,
        shell: Optional[str] = None,
        script_name: Optional[str] = None,
    ) -> None:
        """Initialization.

        Args:
            complete_var (dict): complete arguments dict.
                >>> complete_vars = {
                ...     '-h': 'Display help messages',
                ...     '-v': 'Show version and exit',
                ... }
            script_dir (str): where is the completion file save.
            shell (str, optional): shell type. Defaults to None.
            script_name (str, optional): completion file name. Defaults to None.
            argparse_obj (ArgumentParser, optional): argparse.ArgumentParser. Defaults to None.

        Raises:
            TypeError: when `complete_var` is not dict.
        """
        super(ShellCompletion, self).__init__()

        self.prop = prop
        if not isinstance(complete_vars, dict):
            raise TypeError("complete_var muse be dict.")
        self.complete_vars = complete_vars
        self.script_dir = script_dir
        if not shell:
            shell = self.get_current_shell()
        elif shell.strip() not in self.Supported_Shell:
            raise ValueError(
                "shell name '{0}' is not supported, see {1}".format(
                    shell, self.Supported_Shell
                )
            )
        self.shell = shell
        self.scirpt_name = script_name
        # print(self.complete_var)

    @staticmethod
    def get_current_shell() -> str:
        """Gets the currently used shell.

        Returns:
            (str): Current shell string.
        """
        current_shell = ""
        _, resp = exec_cmd("echo $SHELL")
        if resp:
            current_shell = resp.split("/")[-1].strip()
        return current_shell.lower()

    def generate_resource(self) -> tuple[str, str]:
        """Generate completion scirpt.

        Generate the completion script of the corresponding shell according to
        the template.

        Args:
            shell (str): Current used shell.

        Returns:
            (str): completion file name.
            (str): completion source.
            (str): shell config path.
        """

        if self.shell == "zsh":
            script_name = "zsh_comp"
            template = self._TEMPLATE_ZSH

            def gen_completion():
                vars = []

                for k, desc in self.complete_vars.items():
                    if not desc:
                        desc = "no description."
                    vars.append('    {0}\\:"{1}"\\'.format(k, desc))

                return ("\n".join(vars)).strip()

        elif self.shell == "bash":
            script_name = "bash_comp"
            template = self._TEMPLATE_BASH

            def gen_completion():
                return " ".join(self.complete_vars.keys())

        elif self.shell == "fish":
            script_name = "fish_comp"
            template = self._TEMPLATE_FISH

            def gen_completion():
                return " ".join(self.complete_vars.keys())

        if self.scirpt_name:
            script_name = self.scirpt_name
        complete_content = gen_completion()
        script_src = template % {"prop": self.prop, "complete_vars": complete_content}

        return script_name, script_src

    def write_completion(self, script_name: str, complete_src: str) -> None:
        """Save completion to config path.

        Args:
            name (str): completion name.
            complete_src (str): completion source.

        Returns:
            (str): completion full path.
        """
        # print("\n\n", script_name, "\n\n", complete_src)

        if not os.path.isdir(self.script_dir):
            os.makedirs(self.script_dir, exist_ok=True)
        full_path = os.path.join(self.script_dir, script_name)
        try:
            with open(full_path, "w" if os.path.isfile(full_path) else "x") as f:
                for line in complete_src:
                    f.write(line)
        except Exception as e:
            Log.error(str(e) + str(e.__traceback__))
            return False
        else:
            return True

    def inject_into_shell(self, script_name: str) -> bool:
        """Try using completion script.

        Inject the load of completion script into the configuration of shell.
        If it exists in the configuration, the injection will not be repeated.
        """
        # get shell config path
        _home_ = os.environ["HOME"]
        if self.shell == "zsh":
            config_path = _home_ + "/.zshrc"
        elif self.shell == "bash":
            config_path = _home_ + "/.bashrc"
        elif self.shell == "fish":
            config_path = _home_ + "/.config/fish/config.fish"

        full_script_path = os.path.join(self.script_dir, script_name)
        try:
            with open(config_path) as f:
                shell_conf = f.read()
        except Exception as e:
            raise ShellCompletionError("Read shell config error: {0}".format(e))
        else:
            _re = re.compile(r"{}[^\s]*".format(full_script_path))
            files = _re.findall(shell_conf)

        has_injected = False
        # print(files)
        if files:
            has_injected = True

        if not has_injected:
            try:
                run_cmd('echo "source %s" >> %s ' % (full_script_path, config_path))
            except Exception as e:
                raise ShellCompletionError(f"Inject error: {str(e)}")
            return True
        else:
            return False

    def complete_and_use(self) -> None:
        """Add completion prompt script."""

        print("\n===Try to add completion ...")

        current_shell = self.shell
        print(":: Detected shell: %s" % repr(current_shell))

        # check shell validable.
        if current_shell not in self.Supported_Shell:
            print(":: Don't support completion of %s" % current_shell)
            return None

        # try create completion file.
        script_name, completion_src = self.generate_resource()
        if not self.write_completion(script_name, completion_src):
            print(":: Write completion script failed.")
            return None

        # try inject to shell config.
        try:
            injected = self.inject_into_shell(script_name)
            if injected:
                print(":: Source your shell configuration.")
            else:
                print(":: Command already exist.")
        except Exception as e:
            print(str(e))


def process_argparse(argparse_obj: object) -> dict:
    if not isinstance(argparse_obj, object):
        raise TypeError("Need a argparse.ArgumentParser object.")

    try:
        if not argparse_obj.__dict__.get("_actions", None):
            raise TypeError("Need a argparse.ArgumentParser object.")
    except AttributeError:
        raise TypeError("Need a argparse.ArgumentParser object.")

    arguments = {}

    for action in argparse_obj.__dict__["_actions"]:
        for option in action.option_strings:
            arguments[option] = action.help

    return arguments



# -*- coding:utf-8 -*-

import os
import re
import logging
from typing import Optional
from urllib.request import urlopen

from .utils import confirm
from .common import Fx, Emotion

Log = logging.getLogger(__name__)


class GitignoreGenetor(object):
    """Generate gitignore template.

    Attributes:
        Genres (dict): supported type.

    Raises:
        SystemExit: Can't get template.
        SystemExit: No name.
    """

    # Supported type. https://github.com/github/gitignore
    Supported_Types: dict[str, str] = {
        "android": "Android",
        "c++": "C++",
        "cpp": "C++",
        "c": "C",
        "dart": "Dart",
        "elisp": "Elisp",
        "gitbook": "GitBook",
        "go": "Go",
        "java": "Java",
        "kotlin": "Java",
        "lua": "Lua",
        "maven": "Maven",
        "node": "Node",
        "python": "Python",
        "qt": "Qt",
        "r": "R",
        "ros": "ROS",
        "ruby": "Ruby",
        "rust": "Rust",
        "sass": "Sass",
        "swift": "Swift",
        "unity": "Unity",
    }

    def __init__(self, timeout=60) -> None:
        super(GitignoreGenetor, self).__init__()

        self.timeout = timeout

    def parse_gitignore_page(self, content: str) -> str:
        """Parse html for getting gitignore content.

        Args:
            content (str): template page html.

        Returns:
            (str): gitignore template content.
        """

        # findall table tag, should only one.
        text = re.findall(r"(<table.*?>.*?<\/table>)", content, re.S)
        if not text:
            return ""

        # remove all html tag.
        content_re = re.compile(r"<\/?\w+.*?>", re.S)
        res = content_re.sub("", text[0])
        # replace multi empty line to one line.
        res = re.sub(r"(\n[^\S\r\n]+)+", "\n", res)
        return res

    def get_html_from_url(self, url: str) -> Optional[str]:
        """Crawl gitignore template.

        Args:
            url (str): gitignore template url.

        Returns:
            (str): html string.
        """

        try:
            handle = urlopen(url, timeout=self.timeout)
        except Exception as e:
            Log.error(str(e) + str(e.__traceback__))
            # Exit once an error occurs.
            return None
        else:
            content = handle.read().decode("utf-8")
            return content

    def launch(self, genre: str, dir_path: str) -> None:
        """Try to create gitignore template file.

        Args:
            genre (str): template type, like: 'python'.
            dir_path (str): .gitignore file save path.
        """

        # Process and check the type, and exit in case of any accident.
        name = self.Supported_Types.get(genre.lower(), None)
        if name is None:
            print("Unsupported type: %s" % genre)
            print(
                "Supported type: [{}]. Case insensitive.".format(
                    " ".join(self.Supported_Types.keys())
                )
            )
            return None

        ignore_path = dir_path + "/.gitignore"
        whether_write = True

        # Adjust `.gitignore` whether exist.
        if os.path.exists(ignore_path):
            whether_write = confirm(
                "`.gitignore` existed, overwrite this file? (default: y) [y/n]:"
            )

        if whether_write:
            base_url = "https://github.com/github/gitignore/blob/master/%s.gitignore"
            target_url = base_url % name

            print(
                "Will get ignore file content from %s"
                % (Fx.italic + Fx.underline + target_url + Fx.reset)
            )
            content = self.get_html_from_url(target_url)
            if not content:
                print("Failed to get content and will exit.")
                return None

            ignore_content = self.parse_gitignore_page(content)

            print("Got content, trying to write ... ")
            try:
                with open(ignore_path, "w") as fd:
                    fd.write(ignore_content)
            except Exception as e:
                Log.error(str(e) + str(e.__traceback__))
                print("Write gitignore file failed.")
                print("You can copy it with the following:")
                print("#" * 60)
                print(ignore_content)
            else:
                print(
                    "Write gitignore file successful. {0}".format(Emotion.Icon_Smiler)
                )
